/*
  COPYRIGHT (C) 2003 APPLIED NEUROINFORMATIC GROUP - UNIVERSITY OF BIELEFELD.

  ALL RIGHTS RESERVED.

  REDISTRIBUTION AND USE IN SOURCE AND BINARY FORM, WITH OR WITHOUT
  MODIFICATION, REQUIRE THE PERMISSION OF THE COPYRIGHT HOLDERS.

  COMMERCIAL USE WITHOUT THE EXPLICIT PERMISSION OF THE COPYRIGHT HOLDERS
  IS FORBIDDEN

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS IS''
  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE LIABLE FOR
  ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/



#ifndef CCLUSTERALGORITHM_H
#define CCLUSTERALGORITHM_H

#include "CVector.h"
#include "CDataset.h"
#include "CMetric.h"
#include "CObject.h"
/*!
 * @short Base class for cluster algorithms
 * @author Thorsten Twellmann
 **/
template<class Type>
class CClusterAlgorithm : public CObject<Type>
{
protected:
	CDataset<Type> tReferenceVectors; /**<Dataset with reference vectors of clusters */
	CMetric<Type>* ptMetric;          /**<Metric */
public:

/*!
 * Construct new cluster object that uses the given metric. The algorithm becomes the owner of the metric object
 * and will delete it on destruction.
 **/
	CClusterAlgorithm(CMetric<Type>* ptMetric = new CEuclideanMetric<Type>());

/*!
 * Destructor
 **/
	virtual ~CClusterAlgorithm();

/*!
 * Set metric object to use. The algorithm becomes the owner of the metric object
 * and will delete it.
 *
 * @param ptMetric Pointer to new metric object
 **/
	void setMetric(CMetric<Type>* ptMetric);


/*!
 * Return reference vectors as a dataset
 *
 * @return Dataset of reference vectors
 **/
	CDataset<Type> getReferenceVectors() const;


/*!
 * Return reference to reference vectors as a dataset
 *
 * @return reference to Dataset of reference vectors
 **/
	CDataset<Type>& referenceVectors();

/*!
 * Cluster given dataset
 *
 * @param tDataset Reference to dataset
 **/
	virtual void train(CDataset<Type>& tDataset)=0;


/*!
 * Calculate Root-Mean-Square-Error on given dataset
 * @param tDataset Reference to dataset
 * @return RMSE
 **/

	Type rmse(const CDataset<Type>& tDataset) const;

	/*!
	 * Return index of nearest node
	 **/
	int nearestNode(const CVector<Type>& rtPattern) const;

/*!
 * Return distance to nearest node
 **/
	Type nearestNodeDistance(const CVector<Type>& rtPattern) const;

	/*!
	 * Read/write object to binary stream
	 * @param stream Binary stream
	 * @param mode READ or WRITE
	 * @return False on error
	 **/
	bool serialize(fstream& stream, IO_MODE mode);
};

#endif
