/*
  COPYRIGHT (C) 2003 APPLIED NEUROINFORMATIC GROUP - UNIVERSITY OF BIELEFELD.

  ALL RIGHTS RESERVED.

  REDISTRIBUTION AND USE IN SOURCE AND BINARY FORM, WITH OR WITHOUT
  MODIFICATION, REQUIRE THE PERMISSION OF THE COPYRIGHT HOLDERS.

  COMMERCIAL USE WITHOUT THE EXPLICIT PERMISSION OF THE COPYRIGHT HOLDERS
  IS FORBIDDEN

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS IS''
  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE LIABLE FOR
  ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/



#ifndef CDATASET_H
#define CDATASET_H

#include "CDatasetItem.h"
#include "CDenseVector.h"
#include "CException.h"
#include "CMetric.h"
#include "Macros.h"
#include <vector>
#include <deque>
#include <algorithm>
#include <ctime>
#include <fstream>
#include <iostream>
#include <cstdio>
#include <stdlib.h>
#include <cassert>
#include <cstring>
#include <iomanip>
#include <string>
#include "zlib.h"
#include "gzstream.h"
#include "CObject.h"
#include <ext/algorithm>
#include <cstdlib>
#include <ctime>
using __gnu_cxx::random_sample;


using namespace std;





class RandomNumber {
public:
	RandomNumber() {
		srand(time(0));
	}

	int operator() ( int n) {
		return static_cast<int>( static_cast<float>(n) * rand()/(RAND_MAX+1.0) );
	}
};

/*!
 * @brief Manages pairs of input and output vectors.
 *
 * The class CDataset manages a set of CDatasetItem objects. Each item
 * consists of a pair of CVector objects (input and output vector). There is no need for
 * specifing the exact number of items to store, since memory is allocated automatically.
 * Nevertheless, specifying an approximate number of items or a appropriate value for
 * iGrowSize can speed-up iterative filling of the dataset. When the initial amount of memory
 * is used-up, the dataset will reallocate memory for iGrowSize more items to reduce permanent time
 * consuming reallocations of new memory blocks.
 *
 * The items are stored in a stl vector. Therefore, the same complexity considerations
 * for access and deletion apply to the dataset object:
 *            - Constant time for access of items with getItem(), operator[] or getRandomItem()
 *            - Constant time for removal of the last item  with removeLastItem()
 *            - Linear time for insertion or removal at the beginning or in the middle
 *
 * There are two ways for loading and saving datasets. First, using the operator<< for writing
 * and operator>> for reading data from a stream. These operators can be used e.g. with a fstream to read/write
 * data from a ASCCI file. In this case, the format of the file has to be:
 *              - Number of items, dimension of input and output vector in first line.
 *              - Each following line contains data of one item (input vector, output vector) and a newline command
 *              - Example:
 *                   <code><br>
 *                   21 5 4<br>
 *                   i(0,0) i(0,1) i(0,2) i(0,3) i(0,4) i(0,5) o(0,1) o(0,2) o(0,3) o(0,4)<br>
 *                   i(1,0) i(1,1) i(1,2) i(0,3) i(0,4) i(1,5) o(1,1) o(1,2) o(0,3) o(1,4)<br>
 *                   i(2,0) i(2,1) i(2,2) i(2,3) i(2,4) i(2,5) o(2,1) o(2,2) o(2,3) o(2,4)<br>
 *                   i(x,x) i(x,x) i(x,x) i(x,x) i(x,x) i(x,x) o(x,x) o(x,x) o(x,x) o(x,x)<br>
 *                   i(20,0) i(20,1) i(20,2) i(20,3) i(20,4) i(20,5) o(20,1) o(20,2) o(20,3) o(20,4)<br>
 *                   </code>
 *
 * The second way is to use the method read(char* pcPath), read(istream& istr), write(ostream& ostr) and write(char* pcPath). Using
 * read(char* pcPath) and write(char* pcPath), data will be read/written from/to a gzip-compressed binary file.
 *
 * There are several functions defined in CDatasetAlgorithm.h which modify the input data of a dataset e.g. zscore(..), scaleRange(..), etc.
 **/
template<class Type>
class CDataset : public CObject<Type>

{

protected:
#ifdef DS_IS_A_DEQUE
	deque< CDatasetItem<Type> > tItems;
#else
	vector< CDatasetItem<Type> > tItems;
#endif
	string sType;
	string sInfo;
	int iGrowSize;

	void *pMetaData;



public:

	//! Constructor
	/*!
	 * Constructor. Creates a new dataset object with space reserved for iInitialSize items. Each time the number of items in the
	 * dataset exceeds the reserved size, space for iGrowSize new items will be reserved (the dataset is still empty!). Note: If the dataset is build by repeatedly
	 * appending new items, it is important to reserve an adequate space in order to avoid repeated time-consuming reallocations of memory.
	 * @param iInitialSize Inititially reserved space
	 * @param iGrowSize Increment per reallocation of space for items
	 **/
	CDataset(int iInitialSize=1000, int iGrowSize=1000);

	//! Copy constructor
	/*!
	 * Constructor
	 **/
	CDataset(const CDataset<Type>& rtDataset);


	//! Destructor
	/*!
	 * Destructor
	 **/
	~CDataset();

	bool bProtect;

	//! Assignment operator
	/*!
	 * Assignment operator
	 **/
	CDataset<Type>& operator=(const CDataset<Type>& rtDataset);



	//! Init random seed
	/*!
	 * Init random seed
	 **/
	void initRandomSeed();


	//! Read data set from ascii stream
	/*!
	 * Read dataset from instream
	 * @param istr Stream
	 * @return
	 **/
	bool read(istream& istr);





	virtual string className() const { return string("CDataset");};


	//! Read dataset from binary compressed *.ml file
	/*!
	 * Read dataset from file. The file must be a compressed ml file.
	 * @param pcPath Path to file
	 * @return False in the case of error, true otherwise
	 **/
	bool read(const char* pcPath);


	//! Write dataset to binary and compressed ml file
	/*!
	 * Write dataset to file. The file will be a compressed ml file.
	 * @param pcPath Path to file
	 * @return False in the case of error, true otherwise
	 **/
	bool write(const char* pcPath);


	//! Write dataset to ascii stream
	/*!
	 * Write dataset to outstream
	 * @param ostr Stream
	 * @return
	 **/
	bool write(ostream& ostr);


	bool serialize(fstream& stream, IO_MODE mode = READ);
	bool serialize2(CArchiv& tA);
	//! Set info text
/*!
 * Set info tag
 *@param str Info tag
 **/
	void setInfo(const string& str);


	//! Return info text
	/*!
	 * Get info tag
	 * @return Info tag
	 **/
	string getInfo(void);


	//! Reserve space for iSize items
	void reserve(int iSize);


	//! Set size increment
	void setGrowSize(int iGrowSize);


	//! Split dataset in two disjoint subsets
	/*!
	 * Splits the data in two datasets. The dFraction items (starting from the end) are moved to the new dataset.
	 * @param dFraction Fraction (in percent) to be move to the returned dataset
	 * @return New dataset
	 **/
	CDataset<Type> split(double dFraction);


	//! Copy i random items to a new dataset
	/*!
	 * Returns a random subest of iNumItems. The items remain in the dataset
	 * @param iNbItems Size of subset
	 * @return Subset
	 **/
	CDataset<Type> subset(int iNbItems) const;


	//! Create n disjoint folds
	/*!
	 * Return a vector of iN disjoint and equally sized datasets
	 * @param iN Number of datasets
	 * @return Vector of iN datasets
	 **/
	vector<CDataset<Type> > createFolds(int iN) const ;

	//! Copy or move a range of items to a new dataset
	/*!
	 * Returns the subset starting at element iIndexStart and ending with element iIndexEnd.
	 * If bDelete is true, the corresponding elements will be deleted from the source dataset.
	 * Note that this can be quit slow if iIndexEnd is not the last element of the source set.
	 * @param iIndexStart Index of first element to extract
	 * @param iIndexEnd Index of last element to extract
	 * @param bDelete If true, elements will be deleted from the source set (default: true)
	 * @return Specified subset
	 **/
	CDataset<Type> extract(int iIndexStart, int iIndexEnd, bool bDelete);


	//! Merge a second dataset
	/*!
	 * Append dataset to source dataset
	 * @param Dataset to append
	 **/
	void merge(const CDataset<Type>& rtDataset);

	/*!
	 * (Obsolete! Use item(int) instead) Return item with index iIndex
	 * @return Item
	 **/
	inline const CDatasetItem<Type>& getItem(int iIndex) const;


	//! Return a const reference to item iIndex
	/*!
	 * Return item with index iIndex
	 * @return Item
	 **/
	inline const CDatasetItem<Type>& item(int iIndex) const;


	/*!
	 * Return a const reference to a random item
	 * @return Item
	 **/
	inline const CDatasetItem<Type>& getRandomItem(void) const;


	//! Return the reference on a random item
	/*!
	 * Return random item
	 * @return Item
	 **/
	inline CDatasetItem<Type>& randomItem(void);


	//! Return reference to the last item
	/*!
	 * Return last item
	 * @return Item
	 **/
	CDatasetItem<Type>& getLast() ;


	//! Return reference on item iIndex
	/*!
	 * Return item with index iIndex
	 * @return Item
	 **/
	inline CDatasetItem<Type>& operator[](int iIndex);

	//! Return a const reference to item iIndex
	/*!
	 * Return const item with index iIndex
	 * @return Item
	 **/
	inline const CDatasetItem<Type>& operator()(int iIndex) const;

	/*!
	 *  (Obsolete! Use bestMatchIDInput(..) instead)
	 **/
	int getBestMatchIDInput(const CVector<Type>& rtVector, const CMetric<Type>& rCMetric) const;



	/*!
	 *  Calculates the dataset vector where the input vector is closest to a vector given as an argument.
	 * @param rtVector Reference to a Vector
	 * @param rCMetric Reference to a Metric
	 * @return Index of the best match vector
	 **/
	int bestMatchIDInput(const CVector<Type>& rtVector, const CMetric<Type>& rCMetric) const;

	int bestMatchIDInput(const CVector<Type>& rtVector, const CMetric<Type>& rCMetric,
	                     const int iStart, const int iEnd) const;

	int bestMatchIDInput(const CVector<Type>& rtVector, const CMetric<Type>& rCMetric,
	                     const int iStart, const int iEnd, const int iOmit) const;

	int bestMatchIDInput(const CVector<Type>& rtVector, const CMetric<Type>& rCMetric,
	                     vector<int>::iterator itBegin, vector<int>::iterator itEnd) const;


	/*!
	 * (Obsolete! Use bestMatchIDOutput(..) instead)
	 **/
	int getBestMatchIDOutput(const CVector<Type>& rtVector, const CMetric<Type>& rCMetric) const ;

/*!
 * (Calculates the dataset vector where the output vector is closest to a vector given as an argument.
 * @param rtVector Reference to a Vector
 * @param rCMetric Reference to a Metric
 * @return Index of the best match vector
 **/
	int bestMatchIDOutput(const CVector<Type>& rtVector, const CMetric<Type>& rCMetric) const ;

/*!
 * (Obsolete! Use bestMatchInput(..) instead)
 **/
	void getBestMatchInput(const CVector<Type>& rtVector, const CMetric<Type>& rCMetric, int& riBestID, Type& rtBestDist) const ;
/*!
 * Calculates the dataset vector where the input vector is closest to a vector given as an argument.
 * @param rtVector   Reference to a Vector
 * @param rCMetric   Reference to a Metric
 * @param riBestID   Reference to the ID of the best matching dataset vector (returns the result)
 * @param rtBestDist Reference to the distance between the dataset vector and the argument (returns the result)
 **/
	void bestMatchInput(const CVector<Type>& rtVector, const CMetric<Type>& rCMetric, int& riBestID, Type& rtBestDist) const ;


/*!
 * (Obsolete! Use bestMatchOutput(..) instead)
 **/
	void getBestMatchOutput(const CVector<Type>& rtVector, const CMetric<Type>& rCMetric, int& riBestID, Type& rtBestDist) const ;

	/*!
	 * Calculates the dataset vector where the output vector is closest to a vector given as an argument.
	 * @param rtVector   Reference to a Vector
	 * @param rCMetric   Reference to a Metric
	 * @param riBestID   Reference to the ID of the best matching dataset vector (returns the result)
	 * @param rtBestDist Reference to the distance between the dataset vector and the argument (returns the result)
	 **/
	void bestMatchOutput(const CVector<Type>& rtVector, const CMetric<Type>& rCMetric, int& riBestID, Type& rtBestDist) const ;


	//! Set item iIndex
	/*!
	 * Overwrite item with index iIndex
	 * @param iIndex Index of item to overwrite
	 * @param rtItem Reference to item with new data
	 **/
	int setItem(int iIndex, const CDatasetItem<Type>& rtItem);


	//! Insert item
	/*!
	 * Insert item behind item iIndex
	 * @param iIndex
	 * @param Item to insert
	 **/
	int insertItem(int iIndex, const CDatasetItem<Type>& rtItem);

	/**
	   Sets the entire dataset according to the tData assuming the following
	   data organization: ((input,output),...,(input,output))
	   @param ptData Pointer to the input and output data
	   @param iNbItems Number of items
	   @param iInputDim Input dimension
	   @param iOutputDim Output dimension
	**/
	void getDataFromVector(Type* ptData, int iNbItems, int iInputDim, int iOutputDim);

	/**
	   Sets a vector to the values from a dataset using the following
	   data organization: ((input,output),...,(input,output))
	   @param ptData Pointer to the input and output data
	**/
	void setDataToVector(Type* ptData);



	//! Get input data as matrix
	/*!
	 * Return input data as nxd matrix (item=row)
	 * @return Data matrix
	 **/
	CMatrix<Type> getInputMatrix() const ;


	//! Get output data as matrix
	/*!
	 * Return output data as nxd matrix (item=row)
	 * @return Data matrix
	 **/
	CMatrix<Type> getOutputMatrix() const ;


	//! Append item to the end of the dataset
	/*!
	 * Append item to dataset
	 * @param rtItem Reference to item
	 **/
	int appendItem(const CDatasetItem<Type>& rtItem);


	//! Remove the item iIndex
	/*!
	 * Remove item with index iIndex. Note that removing an item which is not the last item
	 * of the dataset can be quit slow.
	 * @param iIndex Index of item ti remove
	 **/
	int removeItem(int iIndex);



	//! Remove the last item
	/*!
	 * Remove last item
	 **/
	void removeLastItem();


	//! Sort data
	/*!
	 * Sort items
	 **/
	void sortdata() {sort(tItems.begin(),tItems.end());};



	//! Sort data using a generic function as less operator
	/*!
	 * Sort items using a binary comparison function
	 **/
	void sortdata(bool (*func)(CDatasetItem<Type>, CDatasetItem<Type>)) {sort(tItems.begin(),tItems.end(),func);};




	//! Random shuffle of items
	/*!
	 * Random shuffle all items
	 **/
	void randomShuffle();



	//! Delete all items
	/*!
	 * Delete al items
	 **/
	void clear();


	//! Return number of items in the dataset
	/*!
	 * Return current number of items
	 * @return Number of items
	 **/
	inline int items() const { return (int)tItems.size();};


	//! Return dimension of input vectors
	/*!
	 * Return the dimension of the items input vector
	 * @return Input dimension
	 **/
	inline int inputDimension() const;



	//! Return dimension of output vectors
	/*!
	 * Return the dimension of the items input vector
	 * @return Input dimension
	 **/
	inline int outputDimension() const;


	//! Return pointer to meta data object
	/*!
	 * Returns a pointer to a meta data object
	 * @return pointer to meta data
	 **/
	inline void* metaData() { return pMetaData; }


	//! Sets a meta data pointer
	//
	inline void  setMetaData(void* pTheData) { pMetaData = pTheData; }


};



template<class Type>
CDataset<Type>::CDataset(int iInitialSize, int iGrowSize) : CObject<Type>()

{
#ifndef DS_IS_A_DEQUE
	tItems.reserve(iInitialSize);
#endif
	this->iGrowSize=iGrowSize;
	sType    =string("CDataset 0.1");
	sInfo=string("none");
	bProtect = false;
	pMetaData = NULL;
}



template<class Type>
CDataset<Type>::CDataset(const CDataset<Type>& rtDataset): CObject<Type>()

{
	tItems=rtDataset.tItems;
	iGrowSize=rtDataset.iGrowSize;
	sType=rtDataset.sType;
	sInfo=rtDataset.sInfo;
	bProtect = false;
	pMetaData = rtDataset.pMetaData;
}


template<class Type>
CDataset<Type>::~CDataset()
{
	tItems.clear();
}


template<class Type>
void CDataset<Type>::initRandomSeed()
{
	srand(time(NULL));
}

template<class Type>
void CDataset<Type>::clear()
{
	tItems.clear();
}



template<class Type>
CDataset<Type>& CDataset<Type>::operator=(const CDataset<Type>& rtDataset)
{

	if(this == &rtDataset)
		return *this;
	this->clear();
#ifndef DS_IS_A_DEQUE
	this->tItems.reserve(rtDataset.items());
#endif
	this->tItems=rtDataset.tItems;
	return *this;
}


template<class Type>
bool CDataset<Type>::read(const char* pcPath){
	igzstream stream;
	ifstream ist(pcPath, ios::in);

	if(!ist.good())
		return false;
	else
		ist.close();
	stream.open(pcPath, ios::in);

	if(!stream)
		return false;
	read(stream);
	stream.close();
	return true;

}


template<class Type>
bool CDataset<Type>::write(const char* pcPath){
	ogzstream stream;
	stream.open(pcPath, ios::out);

	if(!stream) return false;
	write(stream);
	stream.close();
	return true;

}


template<class Type>
bool CDataset<Type>::read(istream& istr){

	char*  pcBuff;
	float  fBuff;
	int    iBuff;
	int    iNbItems;
	int    iInputDim;
	int    iOutputDim;

	istr.read((char*) &iBuff, sizeof(int));

	pcBuff=new char[iBuff];
	istr.read((char*) pcBuff, iBuff*sizeof(char));

	if(this->sType.compare(pcBuff)){
		cerr << "Wrong file type!" << endl;
		delete[] pcBuff;
		return false;
	}
	delete[] pcBuff;

	istr.read((char*) &iNbItems, sizeof(int));
	istr.read((char*) &this->iGrowSize, sizeof(int));
	istr.read((char*) &iInputDim, sizeof(int));
	istr.read((char*) &iOutputDim, sizeof(int));

	istr.read((char*) &iBuff, sizeof(int));
	pcBuff=new char[iBuff];
	istr.read((char*) pcBuff, iBuff*sizeof(char));
	this->sInfo=pcBuff;
	delete[] pcBuff;

	CDenseVector<Type> inpVec(iInputDim);
	CDenseVector<Type> outVec(iOutputDim);

	this->tItems.clear();
#ifndef DS_IS_A_DEQUE
	this->tItems.reserve(iNbItems);
#endif
	for(int i=0;i<iNbItems;i++){

		for(int j=0;j<iInputDim;j++){
			istr.read((char*) &fBuff, sizeof(float));
			inpVec.setElement(j,static_cast<Type>(fBuff));;
		}

		for(int j=0;j<iOutputDim;j++){
			istr.read((char*) &fBuff, sizeof(float));
			outVec.setElement(j,static_cast<Type>(fBuff));
		}

		this->appendItem(CDatasetItem<Type>(inpVec, outVec, i));
	}
	return true;
}

template<class Type>
bool CDataset<Type>::write(ostream& ostr){

	float fBuff;
	int   iBuff;
	int iNbItems  =this->items();
	int iInputDim =this->inputDimension();
	int iOutputDim=this->outputDimension();

	iBuff=sType.length()+1;
	ostr.write((const char*)&iBuff,sizeof(int));
	ostr.write((const char*)sType.c_str(),iBuff*sizeof(char));
	ostr.write((const char*)&iNbItems,sizeof(iNbItems));
	ostr.write((const char*)&this->iGrowSize,sizeof(this->iGrowSize));
	ostr.write((const char*)&iInputDim,sizeof(iInputDim));
	ostr.write((const char*)&iOutputDim,sizeof(iOutputDim));
	iBuff=sInfo.length()+1;
	ostr.write((const char*)&iBuff,sizeof(int));
	ostr.write((const char*)sInfo.c_str(),iBuff*sizeof(char));

	for(int i=0;i<iNbItems;i++){
		for(int j=0;j<iInputDim;j++){
			fBuff=static_cast<float>(this->getItem(i).getInputComponent(j));
			ostr.write((const char*)&fBuff,sizeof(fBuff));
		}
		for(int j=0;j<iOutputDim;j++){
			fBuff=static_cast<float>(this->getItem(i).getOutputComponent(j));
			ostr.write((const char*)&fBuff,sizeof(fBuff));
		}
	}
	return true;
}

template<class Type>
bool CDataset<Type>::serialize(fstream& stream, IO_MODE mode) {
	if(mode == READ) {
		int iSize;
		stream.read( (char*)&iSize, sizeof(int));
#ifndef DS_IS_A_DEQUE
		tItems.reserve(iSize);
#endif
		stream.read( (char*)&iSize, sizeof(int));
		char* pcBuffer = new char[iSize+1];
		memset(pcBuffer,'\0', iSize+1);
		stream.read( (char*)pcBuffer, iSize*sizeof(char));
		sInfo = string(pcBuffer);
		delete [] pcBuffer;
		CDenseVector<Type> tIn;
		CDenseVector<Type> tOut;
		for(int i=0;i<iSize ;i++) {
			tIn.serialize(stream, mode);
			tOut.serialize(stream, mode);
			tItems.push_back( CDatasetItem<Type>(tIn, tOut) );
		}
	} else {
		int iSize = this->items();
		stream.write( (char*)&iSize, sizeof(int));
		iSize = sInfo.length();
		stream.write( (char*)&iSize, sizeof(int));
		stream.write( (char*)sInfo.c_str(), iSize*sizeof(char));

		for(int i=0;i<iSize ;i++) {
			tItems[i].getInputVector().serialize(stream, mode);
			tItems[i].getOutputVector().serialize(stream, mode);
		}
	}
	return false;
}

template<class Type>
bool CDataset<Type>::serialize2(CArchiv& tA) {
	string str;
	unsigned int uiItems;

	if(!CObject<Type>::serialize2(tA))
		return false;
	if( tA.isReading()) {
		tItems.clear();
		tA >> uiItems;
		tItems.resize(uiItems);
		for(int i=0;i<(int)uiItems ;i++)
			tA >> tItems[i];

		return true;
	} else {
		tA << tItems.size(); tA.flush();
		for(int i=0;i<(int)tItems.size() ;i++)
			tA << tItems[i]; tA.flush();
		return true;
	}
	return false;
}



template<class Type>
void CDataset<Type>::setInfo(const string& str){
	this->sInfo=str;
}

template<class Type>
string CDataset<Type>::getInfo(void){
	return this->sInfo;
}


template<class Type>
void CDataset<Type>::reserve(int iSize){

#ifdef SAFE
	assert(iSize>=0);
#endif
#ifndef DS_IS_A_DEQUE
	tItems.reserve(MAX((int)iSize, (int)tItems.size()));
#endif
}

template<class Type>
void CDataset<Type>::setGrowSize(int iGrowSize){

#ifdef SAFE
	assert(iGrowSize>0);
#endif

	this->iGrowSize=iGrowSize;
}



template<class Type>
inline
CDatasetItem<Type>& CDataset<Type>::operator[](int iIndex)
{
#ifdef SAFE
	assert(iIndex >= 0);
	assert((int)iIndex < (int)tItems.size());
#endif
	return (tItems[iIndex]);
}

template<class Type>
inline
const
CDatasetItem<Type>& CDataset<Type>::operator()(int iIndex) const
{
#ifdef SAFE
	assert(iIndex >= 0);
	assert((int)iIndex < (int)tItems.size());
#endif
	return (tItems[iIndex]);
}

template<class Type>
inline
const CDatasetItem<Type>& CDataset<Type>::getItem(int iIndex) const
{
#ifdef SAFE
	assert(iIndex >= 0);
	assert(iIndex < (int)tItems.size());
#endif
	return (tItems[iIndex]);
}

template<class Type>
inline
const CDatasetItem<Type>& CDataset<Type>::item(int iIndex) const
{
#ifdef SAFE
	assert(iIndex >= 0);
	assert(iIndex < (int)tItems.size());
#endif
	return (tItems[iIndex]);
}


template<class Type>
inline
const CDatasetItem<Type>& CDataset<Type>::getRandomItem(void) const
{
#ifdef SAFE
#endif
	int i = IRAND(0,(int)tItems.size());
	return(tItems[i]);
}

template<class Type>
inline
CDatasetItem<Type>& CDataset<Type>::randomItem(void)
{
#ifdef SAFE
#endif
	int i = IRAND(0,(int)tItems.size());
	return(tItems[i]);
}


template<class Type>
vector<CDataset<Type> > CDataset<Type>::createFolds(int iN) const {
	vector< CDataset<Type> > tVec(iN);

	for(int i=0;i<iN;i++)
		tVec.reserve(this->items());
	for(int i=0;i<items();i++)
		tVec[i%iN].appendItem(getItem(i));
	return tVec;
}

template<class Type>
inline
CDatasetItem<Type>& CDataset<Type>::getLast()
{
	return (tItems[tItems.size()-1]);
}

template<class Type>
inline
int CDataset<Type>::getBestMatchIDInput(const CVector<Type>& rtVector, const CMetric<Type>& rCMetric) const{
	Type tMinDist     = rCMetric.distance(tItems[0].getInputVector(),rtVector);
	int      iBestId  = 0;
	int      iSize    = static_cast<int>(tItems.size());

	for(int i=1;i<iSize;i++){
		Type tDist = rCMetric.distance(tItems[i].getInputVector(),rtVector);
		if(tDist<tMinDist){
			tMinDist=tDist;
			iBestId=i;
		}
	}
	return iBestId;
}

template<class Type>
inline
int CDataset<Type>::bestMatchIDInput(const CVector<Type>& rtVector, const CMetric<Type>& rCMetric) const{
	Type tMinDist     = rCMetric.distance(tItems[0].getInputVector(),rtVector);
	int      iBestId  = 0;
	int      iSize    = static_cast<int>(tItems.size());

	for(int i=1;i<iSize;i++){
		Type tDist = rCMetric.distance(tItems[i].getInputVector(),rtVector);
		if(tDist<tMinDist){
			tMinDist=tDist;
			iBestId=i;
		}
	}
	return iBestId;
}

template<class Type>
inline
int CDataset<Type>::bestMatchIDInput(const CVector<Type>& rtVector, const CMetric<Type>& rCMetric,
                                     const int iStart, const int iEnd) const{
	Type tMinDist     = rCMetric.distance(tItems[iStart].getInputVector(),rtVector);
	int      iBestId  = iStart;

	for(int i=iStart+1; i <= iEnd; i++){
		Type tDist = rCMetric.distance(tItems[i].getInputVector(),rtVector);
		if(tDist<tMinDist){
			tMinDist=tDist;
			iBestId=i;
		}
	}
	return iBestId;
}

template<class Type>
inline
int CDataset<Type>::bestMatchIDInput(const CVector<Type>& rtVector, const CMetric<Type>& rCMetric,
                                     const int iStart, const int iEnd, const int iOmit) const{
	Type tMinDist     = rCMetric.distance(tItems[iStart].getInputVector(),rtVector);

	int iBestId;

	if (iOmit == iStart) {
		iBestId = iStart+1;
	} else {
		iBestId  = iStart;
	}

	for(int i=iStart+1; i <= iEnd; i++) {

		if (i != iOmit) {
			Type tDist = rCMetric.distance(tItems[i].getInputVector(),rtVector);
			if(tDist<tMinDist){
				tMinDist=tDist;
				iBestId=i;
			}
		}
	}
	return iBestId;
}

template<class Type>
inline
int CDataset<Type>::bestMatchIDInput(const CVector<Type>& rtVector, const CMetric<Type>& rCMetric,
                                     vector<int>::iterator itBegin, vector<int>::iterator itEnd) const {

	Type tMinDist = rCMetric.distance(tItems[*itBegin].getInputVector(),rtVector);

	int iBestId = *itBegin;
	vector<int>::iterator it = itBegin + 1;

	for(; it < itEnd; it++) {
		Type tDist = rCMetric.distance(tItems[*it].getInputVector(),rtVector);
		if(tDist<tMinDist){
			tMinDist=tDist;
			iBestId=*it;
		}
	}
	return iBestId;
}

template<class Type>
inline
int CDataset<Type>::getBestMatchIDOutput(const CVector<Type>& rtVector, const CMetric<Type>& rCMetric)const{
	Type tMinDist     = rCMetric.distance(tItems[0].getOutputVector(),rtVector);
	int      iBestId  = 0;
	int      iSize    = static_cast<int>(tItems.size());

	for(int i=1;i<iSize;i++){
		Type tDist = rCMetric.distance(tItems[i].getOutputVector(),rtVector);
		if(tDist<tMinDist){
			tMinDist=tDist;
			iBestId=i;
		}
	}
	return iBestId;
}
template<class Type>
inline
int CDataset<Type>::bestMatchIDOutput(const CVector<Type>& rtVector, const CMetric<Type>& rCMetric)const{
	Type tMinDist     = rCMetric.distance(tItems[0].getOutputVector(),rtVector);
	int      iBestId  = 0;
	int      iSize    = static_cast<int>(tItems.size());

	for(int i=1;i<iSize;i++){
		Type tDist = rCMetric.distance(tItems[i].getOutputVector(),rtVector);
		if(tDist<tMinDist){
			tMinDist=tDist;
			iBestId=i;
		}
	}
	return iBestId;
}



template<class Type>
inline
void CDataset<Type>::getBestMatchInput(const CVector<Type>& rtVector, const CMetric<Type>& rCMetric, int& riBestID, Type& rtBestDist)const{
	int iSize  = static_cast<int>(tItems.size());
	riBestID   = 0;
	rtBestDist = rCMetric.distance(tItems[0].getInputVector(),rtVector);

	for(int i=1;i<iSize;i++){
		Type tDist = rCMetric.distance(tItems[i].getInputVector(),rtVector);
		if(tDist<rtBestDist){
			rtBestDist=tDist;
			riBestID=i;
		}
	}
}

template<class Type>
inline
void CDataset<Type>::bestMatchInput(const CVector<Type>& rtVector, const CMetric<Type>& rCMetric, int& riBestID, Type& rtBestDist)const{
	int iSize  = static_cast<int>(tItems.size());
	riBestID   = 0;
	rtBestDist = rCMetric.distance(tItems[0].getInputVector(),rtVector);

	for(int i=1;i<iSize;i++){
		Type tDist = rCMetric.distance(tItems[i].getInputVector(),rtVector);
		if(tDist<rtBestDist){
			rtBestDist=tDist;
			riBestID=i;
		}
	}
}

template<class Type>
inline
void CDataset<Type>::getBestMatchOutput(const CVector<Type>& rtVector, const CMetric<Type>& rCMetric, int& riBestID, Type& rtBestDist)const{
	int iSize  = static_cast<int>(tItems.size());
	riBestID   = 0;
	rtBestDist = rCMetric.distance(tItems[0].getOutputVector(),rtVector);

	for(int i=1;i<iSize;i++){
		Type tDist = rCMetric.distance(tItems[i].getOutputVector(),rtVector);
		if(tDist<rtBestDist){
			rtBestDist=tDist;
			riBestID=i;
		}
	}
}

template<class Type>
inline
void CDataset<Type>::bestMatchOutput(const CVector<Type>& rtVector, const CMetric<Type>& rCMetric, int& riBestID, Type& rtBestDist)const{
	int iSize  = static_cast<int>(tItems.size());
	riBestID   = 0;
	rtBestDist = rCMetric.distance(tItems[0].getOutputVector(),rtVector);

	for(int i=1;i<iSize;i++){
		Type tDist = rCMetric.distance(tItems[i].getOutputVector(),rtVector);
		if(tDist<rtBestDist){
			rtBestDist=tDist;
			riBestID=i;
		}
	}
}

/**
   Replace item at position iIndex with item given by the pointer
   pItem. If successfull, the dataset is responsible for the
   items memory managment.

   @param iIndex Index of item to replace;
   @param Reference to item.
   @return 1 for success, 0 otherwise.
**/
template<class Type>
int CDataset<Type>::setItem(int iIndex,
                            const CDatasetItem<Type>& tItem)
{
#ifdef SAFE
	assert(iIndex >= 0);
	assert(iIndex < (int)tItems.size());
	assert((int)tItems.size() == 0 || tItems[0].inputDimension() == tItem.inputDimension());
	assert((int)tItems.size() == 0 || tItems[0].outputDimension() == tItem.outputDimension());
#endif
	if(iIndex < 0 || iIndex >= (int)tItems.size())
		return 0;

	tItems[iIndex] = (tItem);
	return 1;
}


/**
   Insert item given by the pointer at position iIndex in the dataset.
   If successfull, the dataset is the new owner and is responsible for
   the items memory managment.


   @param iIndex Position of the item after inserting into the dataset
   @param Reference to item.
   @return 1 for success, 0 otherwise.
**/
template<class Type>
int CDataset<Type>::insertItem(int iIndex,
                               const CDatasetItem<Type>& tItem)
{
#ifdef SAFE
	assert(iIndex >= 0);
	assert(iIndex < (int)tItems.size());
	assert((int)tItems.size() == 0 || tItems[0].inputDimension()  == tItem.inputDimension());
	assert((int)tItems.size() == 0 || tItems[0].outputDimension() == tItem.outputDimension());
#endif
	if(iIndex < 0 || iIndex >= (int)tItems.size())
		return 0;
#ifndef DS_IS_A_DEQUE
	if(tItems.capacity() == tItems.size())
		tItems.reserve((int)tItems.size() + iGrowSize);
#endif
	tItems.insert(tItems.begin()+iIndex, tItem);
	return 1;
}

template<class Type>
void CDataset<Type>::getDataFromVector(Type* ptData, int iNbItems, int iInputDim, int iOutputDim){

#ifdef SAFE
	assert(iNbItems>=0);
	assert(iInputDim>=0);
	assert(iOutputDim>=0);
#endif

	this->clear();

	for(int i=0;i<iNbItems;i++) {
		this->appendItem(CDatasetItem<Type>(CDenseVector<Type>(iInputDim,(ptData+i*(iInputDim+iOutputDim))),
		                                    CDenseVector<Type>(iOutputDim,(ptData+i*(iInputDim+iOutputDim)+iInputDim)),
		                                    i+1));
	}

}

template<class Type>
void CDataset<Type>::setDataToVector(Type* ptData){

	for(int i=0;i<this->items();i++){

		for(int j=0;j<this->inputDimension();j++){
			ptData[i*(this->inputDimension()+this->outputDimension())+j]=this->getItem(i).getInputComponent(j);
		}

		for(int j=0;j<this->outputDimension();j++){
			ptData[i*(this->inputDimension()+this->outputDimension())+this->inputDimension()+j]=this->getItem(i).getOutputComponent(j);
		}

	}

}



template<class Type>
int CDataset<Type>::appendItem(const CDatasetItem<Type>& tItem)
{
#ifdef SAFE
	assert((int)tItems.size() == 0 || tItems[0].inputDimension() == tItem.inputDimension());
	assert((int)tItems.size() == 0 || tItems[0].outputDimension() == tItem.outputDimension());
#endif
#ifndef DS_IS_A_DEQUE
	if(tItems.capacity() == tItems.size())
		tItems.reserve(tItems.size() + iGrowSize);
#endif
	tItems.push_back(tItem);
	return 1;
}


template<class Type>
int CDataset<Type>::removeItem(int iIndex)
{
#ifdef SAFE
	assert(iIndex >= 0);
	assert(iIndex < (int)tItems.size());
#endif
	if(iIndex < 0 || iIndex >=this->items())
		return 0;

	tItems.erase(tItems.begin()+iIndex);
	return 1;
}



template<class Type>
void CDataset<Type>::removeLastItem()
{

	if((items() > 0))
		tItems.erase(tItems.begin()+items()-1);
}


template<class Type>
void CDataset<Type>::randomShuffle()
{
	RandomNumber rnd;
	random_shuffle(tItems.begin(), tItems.end(), rnd);
}




template<class Type>
CDataset<Type> CDataset<Type>::extract(int iIndexStart,
                                       int iIndexEnd,
                                       bool bDelete = true)
{
#ifdef SAFE
	assert(iIndexStart >= 0);
	assert(iIndexStart < (int)tItems.size());
	assert(iIndexEnd >= 0);
	assert(iIndexEnd < (int)tItems.size());
	assert(iIndexStart < iIndexEnd);
#endif
	CDataset<Type> tNewSet;
#ifndef DS_IS_A_DEQUE
	tNewSet.reserve(iIndexEnd-iIndexStart) ;
#endif
	for(int i=iIndexStart;i<=iIndexEnd;i++) {
		tNewSet.appendItem( this->getItem(i) );
	}
	if(bDelete) {
		if(iIndexEnd != (items()-1))
			tItems.erase(this->tItems.begin()+iIndexStart, this->tItems.begin()+iIndexEnd);
		else
			tItems.resize(iIndexStart);
	}
	return tNewSet;
}

template<class Type>
CDataset<Type> CDataset<Type>::split(double dFraction)
{
	CDataset<Type> tNewSet( this->items() );
	int iNumSelect = (int)floor(this->items() * (dFraction/100));
	for(int i=0;i<iNumSelect;i++)
	{
		tNewSet.appendItem( this->getItem(this->items() - 1 - i) );
	}
	tItems.erase(this->tItems.end()-iNumSelect, this->tItems.end());
	return tNewSet;
}

template<class Type>
CDataset<Type> CDataset<Type>::subset(int iNbItems) const {
#ifdef SAFE
	if(iNbItems<0){throw CException("iNbItems<0");}
	if(this->items()<=0){throw CException("this->items()<=0");}

#endif
	ML_LEQ_CHK(iNbItems, tItems.size());
	vector<CDatasetItem<Type> > tTmp(iNbItems);

	random_sample(tItems.begin(), tItems.end(), tTmp.begin(), tTmp.end());

	CDataset<Type> tSubset(iNbItems);
	for(int i=0;i<iNbItems;i++){
		tSubset.appendItem(tTmp[i]);
	}
	return tSubset;

}


template<class Type>
void CDataset<Type>::merge(const CDataset<Type>& rtDataset){
	int iNbItems=rtDataset.items();
#ifndef DS_IS_A_DEQUE
	this->reserve(this->items() + iNbItems);
#endif
	for(int i=0;i<iNbItems;i++){
		this->appendItem(rtDataset.getItem(i));
	}
}

template<class Type>
int CDataset<Type>::inputDimension() const
{
	if((int)tItems.size() == 0) return 0;
	else return tItems[0].inputDimension();
}

template<class Type>
int CDataset<Type>::outputDimension() const
{
	if(tItems.size()==0) return 0;
	else return tItems[0].outputDimension();
}



template<class Type>
CMatrix<Type> CDataset<Type>::getInputMatrix() const {
/*-----------------------------------------------------*/
#ifdef SAVE
	ML_GEQ_CHK(this->items(), 1);
	ML_GEQ_CHK(this->inputDimension(), 1);
#endif
/*-----------------------------------------------------*/

	CMatrix<Type> tData(this->items(), this->inputDimension());

	for(int i=0; i<this->items();i++) {
		tData.setRow(i, tItems[i].getInputVector() );
	}

	return tData;
}



template<class Type>
CMatrix<Type> CDataset<Type>::getOutputMatrix() const {
/*-----------------------------------------------------*/
#ifdef SAVE
	ML_GEQ_CHK(this->items(), 1);
	ML_GEQ_CHK(this->inputDimension(), 1);
#endif
/*-----------------------------------------------------*/

	CMatrix<Type> tData(this->items(), this->inputDimension());

	for(int i=0; i<this->items();i++) {
		tData.setRow(i, tItems[i].getOutputVector() );
	}
	return tData;
}











#endif
