/*
  COPYRIGHT (C) 2003 APPLIED NEUROINFORMATIC GROUP - UNIVERSITY OF BIELEFELD.

  ALL RIGHTS RESERVED.

  REDISTRIBUTION AND USE IN SOURCE AND BINARY FORM, WITH OR WITHOUT
  MODIFICATION, REQUIRE THE PERMISSION OF THE COPYRIGHT HOLDERS.

  COMMERCIAL USE WITHOUT THE EXPLICIT PERMISSION OF THE COPYRIGHT HOLDERS
  IS FORBIDDEN

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS IS''
  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE LIABLE FOR
  ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/



#ifndef CDATASETALGORITHM_H
#define CDATASETALGORITHM_H

#include "CDenseVector.h"
#include "CMatrix.h"
#include "CMetric.h"
#include "CDataset.h"
#include "CDatasetItem.h"
#include <vector>
#include <algorithm>
#include <ctime>
#include <iostream>
#include "CObject.h"




/*!
 * Find the maximal element of each component of the input vectors
 * @param Dataset
 * @return Vector with maximal elements
 **/
template <class Type>
CDenseVector<Type> minElements(const CDataset<Type>& tDs);

/*!
 * Find the minimal element of each component of the input vectors
 * @param Dataset
 * @return Vector with minimal elements
 **/
template <class Type>
CDenseVector<Type> maxElements(const CDataset<Type>& tDs);


/*!
 * Normalizes data to zero mean
 * @param tDs Dataset
 * @param tMean To be filled with subracted mean
 * @return Centered dataset
 **/
template <class Type>
CDataset<Type> centerInput(const CDataset<Type>& tDs, CDenseVector<Type>& tMean);

/*!
 * Normalizes data to zero mean and zero variance
 * @param tDs Dataset
 * @param tMean To be filled with subracted mean
 * @param tVariance To be filled with variance
 * @return Centered dataset
 **/
template <class Type>
CDataset<Type> zscore(const CDataset<Type>& tDs, CDenseVector<Type>& tMean, CDenseVector<Type>& tVariance);

/*!
 * Normalizes data to zero mean and zero variance. In contrast to  zscore(..), this function uses the mean and variance
 * given by the function's parameters.
 *
 * @param tDs Dataset
 * @param tMean Mean vector
 * @param tVariance Variance vector
 * @return dataset
 **/
template <class Type>
CDataset<Type> applyZScore(const CDataset<Type>& tDs, const CDenseVector<Type>& tMean, const CDenseVector<Type>& tVariance);

/*!
 * Normalizes data to zero mean and zero variance. In contrast to  zscore(..), this function uses the mean and variance
 * given by the function's parameters.
 *
 * @param tDs Vector
 * @param tMean Mean vector
 * @param tVariance Variance vector
 * @return dataset
 **/
template <class Type>
CDenseVector<Type> applyZScore(const CDenseVector<Type>& tDs, const CDenseVector<Type>& tMean, const CDenseVector<Type>& tVariance);


/*!
 * Scales each component of the input data to the range given by tNewMin and tNewMax
 *
 * @param tDs Dataset
 * @param tNewMin New min values of the input components
 * @param tNewMax New max values of the input components
 * @param tShift To filled with shift vector
 * @param tScale To filled with scale vector
 * @return dataset
 **/
template <class Type>
CDataset<Type> scaleRange(const CDataset<Type>& tDs, const CDenseVector<Type> tNewMin,  const CDenseVector<Type> tNewMax,CDenseVector<Type>& tShift, CDenseVector<Type>& tScale);

/*!
 * Scales each component of the input data to the range given by tNewMin and tNewMax. In contrast to scaleRange(..), this function
 * uses shift and scale given by the functions parameters
 *
 * @param tDs Dataset
 * @param tShift Shift vector
 * @param tScale Scale vector
 * @return dataset
 **/
template <class Type>
CDataset<Type> applyScaleRange(const CDataset<Type>& tDs, const CDenseVector<Type>& tShift, const CDenseVector<Type>& tScale);


/*!
 * Scales each component of the input data to the range given by tNewMin and tNewMax. In contrast to scaleRange(..), this function
 * uses shift and scale given by the functions parameters
 *
 * @param tDs Dataset
 * @param tShift Shift vector
 * @param tScale Scale vector
 * @return dataset
 **/
template <class Type>
CDenseVector<Type> applyScaleRange(const CDenseVector<Type>& tDs, const CDenseVector<Type>& tShift, const CDenseVector<Type>& tScale);

/**
   Normalizes the length of each (input) vector according to a
   given metric (using a vector with zero components).
   @param rtDataset Reference to a dataset object.
   @param rtMetric Reference to a metric object.
**/
template <class Type>
CDataset<Type> normalize(const CDataset<Type>& rtDs, const CMetric<Type>& rtMetric);

/**
   Removes the projection of rtVec from all vectors in the dataset.
   @param rtDataset Reference to a dataset object.
   @param rtVec     Reference to a the vector which should be eliminated
**/
template <class Type>
CDataset<Type>  removeProjection(const CDataset<Type>& rtDataset, const CDenseVector<Type>& rtVec);

/**
   Apply a weight vector by scaling each (input) component with desired factor.
   @param rtDataset  Reference to a dataset object.
   @param rtWeights  weight vector.
**/
template <class Type>
void applyWeights(CDataset<Type>& rtDataset, const CDenseVector<Type>& rtWeights);

/**
   Normalizes the euclidean length of each (input) vector
**/
template <class Type>
void normalizeInputDataEuclidean(CDataset<Type>& rtDs);




#endif
