/*
  COPYRIGHT (C) 2003 APPLIED NEUROINFORMATIC GROUP - UNIVERSITY OF BIELEFELD.

  ALL RIGHTS RESERVED.

  REDISTRIBUTION AND USE IN SOURCE AND BINARY FORM, WITH OR WITHOUT
  MODIFICATION, REQUIRE THE PERMISSION OF THE COPYRIGHT HOLDERS.

  COMMERCIAL USE WITHOUT THE EXPLICIT PERMISSION OF THE COPYRIGHT HOLDERS
  IS FORBIDDEN

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS IS''
  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE LIABLE FOR
  ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/



#ifndef CDATASETITEM_H
#define CDATASETITEM_H

#include "CVector.h"
#include "CDenseVector.h"
#include "assert.h"
#include "stdio.h"
#include "CObject.h"
#include "CArchiv.h"
#include "CObjectFactory.h"
/*!
 * @brief Single item of a dataset consisting of a pair of input and out vectors.
 * @author Thorsten Twellmann
 **/


template<class Type>
class CDatasetItem :public CObject<Type> {
protected:
	CVector<Type>* ptInput;
	CVector<Type>* ptOutput;
	int iId;

public:
	CDatasetItem()
	{
		ptInput  = new CDenseVector<Type>(1);
		ptOutput = new CDenseVector<Type>(1);
		iId=-1;
	};

	CDatasetItem(int iDimInput, int iDimOutput) : CObject<Type>() {
#ifdef SAFE
		assert(iDimInput >= 0);
		assert(iDimOutput >= 0);
#endif
		ptInput  = new CDenseVector<Type>(iDimInput);
		ptOutput = new CDenseVector<Type>(iDimOutput);
		iId=-1;
	};


	CDatasetItem(const CVector<Type>& rtInput,const CVector<Type>& rtOutput) : CObject<Type>()
	{
		ptInput  = rtInput.clone();
		ptOutput = rtOutput.clone();
		iId=-1;
	};

	CDatasetItem(const CVector<Type>& rtInput,const CVector<Type>& rtOutput,const int iID) : CObject<Type>()
	{
		ptInput  = rtInput.clone();
		ptOutput = rtOutput.clone();
		iId = iID;
	};

	CDatasetItem(const CDatasetItem<Type>& rtItem) : CObject<Type>()
	{
		this->ptInput = rtItem.ptInput->clone();
		this->ptOutput= rtItem.ptOutput->clone();
		this->iId = rtItem.iId;
	};

	virtual string className() const { return string("CDatasetItem");};
	virtual CDatasetItem<Type>& operator=(const CDatasetItem<Type>& rtItem)
	{
		if(this == &rtItem)
			return *this;
		delete this->ptInput;
		delete this->ptOutput;
		this->ptInput = rtItem.ptInput->clone();
		this->ptOutput= rtItem.ptOutput->clone();
		this->iId = rtItem.iId;
		return *this;
	};


	/** Initializes CDatasetItem from given pointer arguments.
	    Assumes ownership and will delete them on destruction. */
	CDatasetItem(CVector<Type>* in, CVector<Type>* out)
	{
		ptInput = in;
		ptOutput = out;
		iId=-1;
	}

	virtual CDatasetItem<double>* convertToDouble()
	{
		if(sizeof(Type) == sizeof(double))
			return new CDatasetItem<double>(*((CDatasetItem<double>*)this));

		return new CDatasetItem<double>(ptInput->toDouble(),
		                                ptOutput->toDouble());
	};


	virtual CDatasetItem<float>* convertToFloat()
	{
		if(sizeof(Type) == sizeof(float))
			return new CDatasetItem<float>(*((CDatasetItem<float>*) this));

		return new CDatasetItem<float>(ptInput->toFloat(),
		                               ptOutput->toFloat());

	}

	virtual ~CDatasetItem() {
		if(ptInput)
			delete ptInput;
		if(ptOutput)
			delete ptOutput;
	};


	void setInputVector(const CVector<Type>& rtVector)
	{
		if(ptInput) {
			delete ptInput;
		}
		ptInput = rtVector.clone();
	};


	void setOutputVector(const CVector<Type>& rtVector)
	{
		if(ptOutput) {
			delete ptOutput;
		}
		ptOutput = rtVector.clone();
	};


	/*!
	 * Obsolete! Use inputVector(..) instead
	 **/
	inline const CVector<Type>& getInputVector() const
	{
		return *ptInput;
	};

	inline const CVector<Type>& inputVector() const
	{
		return *ptInput;
	};


	/*!
	 * Obsolete! Use inputVector(..) instead
	 **/
	inline CVector<Type>& getInputVector()
	{
		return *ptInput;
	};

	inline CVector<Type>& inputVector()
	{
		return *ptInput;
	};


	inline CVector<Type>* inputVectorPtr() const {
		return ptInput;
	}


	/*!
	 * Obsolete! Use outputVector(..) instead
	 **/
	inline const CVector<Type>& getOutputVector() const
	{
		return *ptOutput;
	}
	inline const CVector<Type>& outputVector() const
	{
		return *ptOutput;
	}
	/*!
	 * Obsolete! Use outputVector(..) instead
	 **/
	inline CVector<Type>& getOutputVector()
	{
		return *ptOutput;
	};
	inline CVector<Type>& outputVector()
	{
		return *ptOutput;
	};

	inline CVector<Type>* outputVectorPtr() const {
		return ptOutput;
	}

	/*!
	 * Obsolete! Use intputComponent(..) instead
	 **/
	inline Type getInputComponent(int iIndex) const
	{
		return ptInput->getElement(iIndex);
	};
	inline Type inputComponent(int iIndex) const
	{
		return ptInput->getElement(iIndex);
	};

	/*!
	 * Obsolete! Use outputComponent(..) instead
	 **/
	inline Type getOutputComponent(int iIndex) const
	{
		return ptOutput->getElement(iIndex);
	};
	inline Type outputComponent(int iIndex) const
	{
		return ptOutput->getElement(iIndex);
	};

	/*    inline void setInputComponent(int iIndex, Type tValue)
	      {
	      ptInput->setElement(iIndex, tValue);
	      };


	      inline void setOutputComponent(int iIndex, Type tValue)
	      {
	      ptOutput->setElement(iIndex, tValue);
	      };    */

	inline int inputDimension() const
	{
		return ptInput->dimension();
	};


	inline int outputDimension() const
	{
		return ptOutput->dimension();
	};

	virtual inline bool  operator== (const CDatasetItem<Type>& right) const
	{
		if((int)this->id() == (int)right.id())
			return 1;
		else
			return 0;
	};

	virtual inline bool operator!= (const CDatasetItem<Type>& right) const
	{
		if((int)this->id() != (int)right.id())
			return 1;
		else
			return 0;
	};

	virtual inline bool  operator> (const CDatasetItem<Type>& right) const
	{
		if((int)this->id() < (int)right.id())
			return 0;
		else
			return 1;
	};


	virtual inline bool  operator< (const CDatasetItem<Type>& right) const
	{
		if((int)this->id() > (int)right.id())
			return 0;
		else
			return 1;
	};

	inline int id() const  {return iId;};

	inline void setId(int iId) { this->iId = iId;};


	/*!
	 * Read/write object from/to binary stream
	 * @param stream Reference to stream
	 * @param Mode (READ/WRITE)
	 * @return False on error
	 **/
	bool serialize(fstream& stream, IO_MODE mode) {
		CObject<Type>::serialize(stream, mode);
		if(mode == WRITE) {
			stream.write((char*) &iId, sizeof(int));
			stream << ptInput->className() << endl;
			ptInput->serialize(stream, mode);
			stream << ptOutput->className() << endl;
			ptOutput->serialize(stream, mode);
			return true;
		} else {
			string str;
			stream.read((char*) &iId, sizeof(int));
			getline(stream, str);
			ptInput = (CVector<Type>*) CObjectFactory::instance()->produceObject(str, this->dataType());
			ptInput->serialize(stream, mode);
			getline(stream, str);
			ptOutput = (CVector<Type>*) CObjectFactory::instance()->produceObject(str, this->dataType());
			ptOutput->serialize(stream, mode);
			return true;
		}
		return false;
	}


	bool serialize2(CArchiv& tA) {
		string str;
		if(!CObject<Type>::serialize2(tA))
			return false;
		if( tA.isReading()) {
			if(ptInput)
				delete ptInput;
			if(ptOutput)
				delete ptOutput;
			tA >> iId;
			tA >> str;
			ptInput = (CVector<Type>*) CObjectFactory::instance()->produceObject(str, this->dataType());
			if(!ptInput)
				ML_CRITICAL_MSG("Unable to produce CVector object");
			tA >> *ptInput;
			tA >> str;
			ptOutput = (CVector<Type>*) CObjectFactory::instance()->produceObject(str, this->dataType());
			if(!ptOutput)
				ML_CRITICAL_MSG("Unable to produce CVector object");
			tA >> *ptOutput;
			return true;
		} else {
			tA << iId; tA.flush();
			tA << ptInput->className(); tA.flush();
			tA << *ptInput; tA.flush();
			tA << ptOutput->className(); tA.flush();
			tA << *ptOutput; tA.flush();
			return true;
		}
		return false;
	}


};
#endif
