/*
  COPYRIGHT (C) 2003 APPLIED NEUROINFORMATIC GROUP - UNIVERSITY OF BIELEFELD.

  ALL RIGHTS RESERVED.

  REDISTRIBUTION AND USE IN SOURCE AND BINARY FORM, WITH OR WITHOUT
  MODIFICATION, REQUIRE THE PERMISSION OF THE COPYRIGHT HOLDERS.

  COMMERCIAL USE WITHOUT THE EXPLICIT PERMISSION OF THE COPYRIGHT HOLDERS
  IS FORBIDDEN

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS IS''
  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE LIABLE FOR
  ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/



#ifndef CDATASETSTATISTIC_H
#define CDATASETSTATISTIC_H

#include "CDenseVector.h"
#include "CMatrix.h"
#include "CMetric.h"
#include "CDataset.h"
#include "CDatasetItem.h"
#include <vector>
#include <algorithm>
#include <ctime>
#include <iostream>
#include "CObject.h"
/******************************************************************************/
/**
   Template class for CDatasetStatistics.
**/
template<class Type>
class CDatasetStatistic: public CObject<Type>{
private:
	CDataset<Type>* ptData;

public:

/*----------------------------------------------------------------------------*/
	/**
	   Constructor
	**/
	CDatasetStatistic(void);

/*----------------------------------------------------------------------------*/
	/**
	   Constructor
	**/
	CDatasetStatistic(CDataset<Type>* ptDataset);

/*----------------------------------------------------------------------------*/
	/**
	   Return the input dimesion
	   @return input dimension
	**/
	int inputDimension(void);



	//! Calculate mean vector of elements in rtDs with class label tClassLabel
	static CDenseVector<Type> mean(const CDataset<Type>& rtDs, Type tClassLabel);


	//! Calculate mean vector of elements in rtDs
	static CDenseVector<Type> mean(const CDataset<Type>& rtDs);


	//! Calculate covariance matrix of elements in rtDs with class label tClassLabel
	static CMatrix<Type> covarianceMatrix(const CDataset<Type>& rtDs, Type tClassLabel);


	//! Calculate covariance matrix of elements in rtDs
	static CMatrix<Type> covarianceMatrix(const CDataset<Type>& rtDs);

/*----------------------------------------------------------------------------*/
	/**
	   Sets the dataset
	   @param ptSet Pointer to the new dataset
	**/
	void setData(CDataset<Type>* ptSet);

	virtual string className() const { return string("CDatasetStatistic");};
/*----------------------------------------------------------------------------*/
	/**
	   Gets the data
	   @return copy of the internal dataset
	**/
	CDataset<Type>* getData(void);

/*----------------------------------------------------------------------------*/
	/**
	   Gets the auto-correlation matrix.
	   @return tACMatrix Auto-correlation matrix of the input data.
	**/
	CMatrix<Type> getAutoCorrelation(void);

/*----------------------------------------------------------------------------*/
	/**
	   Gets the class independet LDA-matrix.
	   (inv(within-class scatter)*between class scatter).
	**/
	CMatrix<Type> getLDA(void);

/*----------------------------------------------------------------------------*/
	/**
	   Gets the whitin class scatter matrix.
	**/
	CMatrix<Type> getSW(void);

/*----------------------------------------------------------------------------*/
	/**
	   Gets the between class scatter matrix.
	**/
	CMatrix<Type> getSB(void);

/*----------------------------------------------------------------------------*/
	/**
	   Calculates the maximal value for each feature of the input vector.
	   @return Vector with maximal values.
	**/
	CDenseVector<Type> maxElements();

/*----------------------------------------------------------------------------*/
	/**
	   Calculates the minimal value for each feature of the input vector.
	   @return Vector with minimal values.
	**/
	CDenseVector<Type> minElements();


	/**
	   Calculates several statistics for the input vectors. Each statistical value
	   corresponds to one feature of the input vectors.

	   @param rtMean Vector which will be filled with mean values
	   @param rtDeviation Vector which will be filled with standard deviation
	   @param rtVariance Vector which will be filled with variance
	**/
	void statistic(CDenseVector<Type>& rtMean,
	               CDenseVector<Type>& rtDeviation,
	               CDenseVector<Type>& rtVariance) const;


	/**
	   Same as above, but for calculation of the statistical values, only items with output vector
	   equal to rtPattern are considered.

	   @param rtPattern Pattern
	   @param rtMean Vector which will be filled with mean values
	   @param rtDeviation Vector which will be filled with standard deviation
	   @param rtVariance Vector which will be filled with variance
	   @param Number of items considered
	**/
	int statistic(const CDenseVector<Type>& rtPattern,
	              CDenseVector<Type>& rtMean,
	              CDenseVector<Type>& rtDeviation,
	              CDenseVector<Type>& rtVariance) const;

	/**
	   Computes for each element a ranked list of distances to the other elements.
	   Uses the input vectors for the computation of the distances.
	   Creates a new dataset where for each item,
	   -- the input vector contains the ranked list of the IDs and
	   -- the output vector the corresponding list of distances

	   @param iSize: number of dataset elements to consider for rank matrix
	   @param rMetric: metric to use for computation of distance in input space
	**/
	CDataset<float>* rankInputSpace(int iSize, CMetric<Type>& rMetric);

	/**
	   Similar to above, but use the distances in the map space.
	   Additionally expects the distances of each data item to the prototype vectors
	   (used for better sorting of distances in map space)
	   Creates a new dataset where for each item,
	   -- the input vector contains the ranked list of the IDs and
	   -- the output vector the corresponding list of distances

	   Expects the following data structure:
	   this->ptData input vectors:  vectors in input space (unused)
	   output vectors: component 0: ID of best match unit in SOM
	   1: distance to prototype vector

	   rNodeCoords input vectors: coordinates of nodes in map space
	   output vectors: unused

	   @param iSize: number of dataset elements to consider for rank matrix
	   @param rNodeCoords: node coordinates in map space
	   @param rMetricMap: metric to compute node distances in map space
	**/
	CDataset<float>* rankMapSpace(int iSize,
	                              CDataset<double>& rNodeCoords,
	                              CMetric<double>& rMetricMap);

	/**
	   Computes Spearmann's rho

	   Expects the following data structure
	   this->ptData input vectors:  vectors in input space
	   output vectors: component 0: ID of best match unit in SOM
	   1: distance to prototype vector

	   rNodeCoords input vectors: coordinates of nodes in map space
	   output vectors: unused

	   @param iSize: number of dataset elements to consider for rank list
	   @param rNodeCoords: node coordinates in map space
	   @param rMetricMap: metric to compute node distances in map space
	   @param rMetricData: metric to compute distances in feature space
	**/
	float spearman(int iSize,
	               CDataset<double>& rNodeCoords,
	               CMetric<double>& rMetricMap,
	               CMetric<Type>& rMetricData);

	/**
	   Same as above, but data will be dumped to stdout
	**/
	void dumpStatistic();

	/*----------------------------------------------------------------------------*/
	/**
	   Default destructor.
	**/
	~CDatasetStatistic();
};
#endif
