/*
  COPYRIGHT (C) 2003 APPLIED NEUROINFORMATIC GROUP - UNIVERSITY OF BIELEFELD.

  ALL RIGHTS RESERVED.

  REDISTRIBUTION AND USE IN SOURCE AND BINARY FORM, WITH OR WITHOUT
  MODIFICATION, REQUIRE THE PERMISSION OF THE COPYRIGHT HOLDERS.

  COMMERCIAL USE WITHOUT THE EXPLICIT PERMISSION OF THE COPYRIGHT HOLDERS
  IS FORBIDDEN

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS IS''
  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE LIABLE FOR
  ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/



#ifndef CDENSEVECTOR_H
#define CDENSEVECTOR_H

#include "CVector.h"
#include "CDenseVector.h"
#include "CMatrix.h"
#include "Macros.h"


template<typename Type> struct keyed_iterator;

template<class Type> class CDenseVector;
template<class Type> class CMatrix;

/**
   Elementwise multiplication of two vectors
**/
template<class Type> CDenseVector<Type> elementMul(const CDenseVector<Type>& rtVec1, const CDenseVector<Type>& rtVec2);

/**
   Outer product  of two vectors
**/
template<class Type> CMatrix<Type> outerProduct(const CDenseVector<Type>& rtVec1, const CDenseVector<Type>& rtVec2);

/*!
 *  @brief Template object for vectors of single and double precision and integer.
 **/
template<class Type>
class CDenseVector :  public CVector<Type>
{
protected:
	int iDim;
	Type* ptData;

	void copy(const CVector<Type>&vec);

public:
	/**
	   Constructor -- Constructs a new instance of a vector object.
	   @param iDim Dimension of vector.
	**/
	explicit CDenseVector(int iDim = 0);

	/**
	   Constructor -- Constructs a new instance of a vector object.
	   @param iDim Dimension of vector (default 1).
	   @param ptData Array of dimension iDim with initial data
	**/
	CDenseVector(int iDim, const Type* ptData);

	/**
	   Constructor -- Constructs a new instance of a vector object. Each element
	   is initialized with value of tValue.
	   @param iDim Dimension of vector (default 1).
	   @param ptData Array of dimension iDim with initial data (default: NULL)
	**/
	CDenseVector(int iDim, Type tValue );

	/**
	   Constructor -- Constructs a new vector instance of a vector.
	   The vector is initialized with the values from a stl vector.
	   @param rtVec Reference on a stl vector to clone.
	**/
	CDenseVector(const vector<Type>& rtVec);

	/**
	   Constructor -- Constructs a new vector instance which is an exact clone of
	   a second vector.
	   @param tVec Reference on vector to clone.
	**/
	CDenseVector(const CDenseVector<Type>& tVec);

	CDenseVector(const CVector<Type>& vec);
	virtual string className() const { return string("CDenseVector");};
	/**
	   Destructor
	**/
	virtual ~CDenseVector();

	CVector<Type>* clone() const;

	/**
	   Initializes all vector elements with 0.
	**/
	void zeros(void);

	/**
	   Initializes all vector elements with 1.
	**/
	void ones(void);

	/**
	   Returns copy of vector using double values
	**/

	CVector<double>* toDouble() const;


	/**
	   Returns copy of vector using float values
	**/

	CVector<float>* toFloat() const;

	vector<Type> toSTL() const;

	CDenseVector<int> sort();



	/**
	   Assignment operator -- Assigns a vector instance A to a
	   varible B. The dimension of B does not have to match the
	   dimension of B, since the size of B is adjusted according A.

	   @param Reference on vector B.
	**/
	CDenseVector<Type>& operator=(const CDenseVector<Type>& tSrc);

	/**
	   Returns the dimension of the vector.

	   @return Dimension of vector.
	**/
	int dimension() const { return iDim; };

	/**
	   Returns vector as dim x 1 matrix.

	   @return Dimension of vector.
	**/
	CMatrix<Type> asMatrix() const;

	/**
	   Returns the element of the vector at position iIndex. The index of the
	   vector elements starts with 0.

	   @return Value of element at position iIndex.
	**/
	Type getElement(int iIndex) const;

	const Type& operator[](int i) const {
		ML_LS_CHK(i,iDim);
		return ptData[i];
	}

	/**
	   Returns a reference on the element of the vector at position iIndex. The index of the
	   vector elements starts with 0.

	   @return Value of element at position iIndex.
	**/
	Type& operator[](int i) {
		ML_LS_CHK(i,iDim);
		return ptData[i];
	};

	/**
	   Assigns the element at postion iIndex a new value.

	   @param iIndex Index of element.
	   @param tValue New value of element.
	**/
	void setElement(int iIndex, Type tValue);

	/**
	   Resizes dimension of vector. If the new vector is smaller, elements
	   at the end of the vector are removed according the new size. If the
	   new vector is larger, elements with value zero are appended to the
	   existing vector data.

	   @param iNewDim New vector dimension
	**/
	void resize(int iNewDim);

	/**
	   Adds a scalar to each vector element.

	   @param tScalar
	**/
	void operator+=(Type tScalar);

	/*!
	 * Checks for NaN elements
	 * @return True if one or more elements are NaN, false otherwise
	 **/
	bool isNaN() const;


	CVector<Type>& operator*=(const CVector<Type>& v2) {
		CVector<Type>::operator*=(v2);
		return *this;
	}
	CVector<Type>& operator/=(const CVector<Type>& v2) {
		CVector<Type>::operator/=(v2);
		return *this;
	}

	/// multiply every element with scalar
	virtual CVector<Type>& operator*=(const Type&);
	/// divide every element by scalar
	virtual CVector<Type>& operator/=(const Type&);

	/**
	   Subtract scalar from each vector element.

	   @param tScalar
	**/
	void operator-=(Type tScalar);

	/**
	   Adds two vectors.

	   @param tVec
	**/
	CDenseVector<Type>& operator+=(const CVector<Type>& tVec);

	/**
	   Subtracts two vectors.

	   @param tVec
	**/
	CDenseVector<Type>& operator-=(const CVector<Type>& tVec);

	/**
	   Calculates the manhatten distance between two vectors.

	   @param Reference to vector instance B.
	   @return Distance.
	**/
	Type manhattenDistance( const CDenseVector<Type>& rtVec) const;

	/**
	   Calculates the maximum distance between two vectors.

	   @param Reference to vector instance B.
	   @return Distance.
	**/
	Type maximumDistance( const CDenseVector<Type>& rtVec) const;

	/*!
	  Sum of elements
	**/
	Type sumElements() const;

	/*!
	  Append vector
	**/
	void cat(const CDenseVector<Type>& tVec);

	/**
	   Returns the smallest value of the vector elements.

	   @return Minimal value of elements.
	**/
	Type minElement()  const;

	/**
	   Returns the greatest value of the vector elements.

	   @return Maximal value of elements.
	**/
	Type maxElement() const;

	/**
	   Returns the index of the element containing the smallest value.

	   @return Index of element.
	**/
	int minElementIndex() const;

	/**
	   Returns the index of the element containing the smallest value.

	   @return Index of element.
	**/
	int maxElementIndex() const;




	/**
	   Return vector of elementwise squareroot.

	   @return Vector.
	**/
	CDenseVector<Type> squareroot() const;

	/**
	   Return vector of elementwise power.

	   @return Vector.
	**/
	CDenseVector<Type> power(int iPower = 2) const;


	/**
	   Returns the vector of componentwise maximum values of two vectors.

	   @return Maxima vector.
	**/
	CDenseVector<Type> maxElements( const CDenseVector<Type>& rtVec) const;

	/**
	   Returns the vector of componentwise maximum values of two vectors.

	   @return Maxima vector.
	**/
	CDenseVector<Type> minElements( const CDenseVector<Type>& rtVec) const;

	/**
	   Return a subvector containing iElements elements starting from iStart (inclusive iStart)
	   @return Subvector.
	**/
	CDenseVector<Type>  subVector(int iStart, int iElements) const;

	/**
	   Returns the mean value of the vector elements.

	   @return Mean value.
	**/
	Type mean()  const;



	/**
	   Returns the standard deviation of the vector elements.

	   @return Standard deviation value.
	**/
	Type stdDev()  const;
	/**
	   Returns a new vector contraining the elements of the current vector shifted
	   by iNumElements to the left. Elements pushed out at the left end will be  inserted at the
	   right end.

	   @param Number of elements to shift.
	   @return Shifted vector.
	**/
	CDenseVector<Type> shiftLeft(int iNumElements=1);

	/**
	   Returns a new vector contraining the elements of the current vector shifted
	   by iNumElements to the right. Elements pushed out at the right end will be inserted at the
	   left end.

	   @param Number of elements to shift.
	   @return Shifted vector.
	**/
	CDenseVector<Type> shiftRight(int iNumElements=1);

	/**
	   Scales each element according the give intervall edition. If the current value of element i is in
	   the interval tCurrentMin_i to tCurrentMax_i, the new element value is calculated by scaling tCurrentMin, tCurrentMax
	   to tNewMin, tNewMax.

	   @param tCurrentMin, tCurrentMax Min/max values of the intervals from which each element value can be taken.
	   @param tNewMin, tNewMax New min/max values for the elementwise intervals.
	   @return Vector scaled according the new bmin/max values.
	**/
	CDenseVector<Type> scale(const CDenseVector<Type>& tCurrentMin, const CDenseVector<Type>& tCurrentMax,
	                                const CDenseVector<Type>& tNewMin, const CDenseVector<Type>& tNewMax) const ;



	/**
	   Import vector named acVarname from Matlab MAT-formated file acPath
	**/
	int importFromMatFile(const char* acPath, const char* acVarname);
	/**
	   Save vector with name acVarname in the MAT-formated file acPath. if bAppend
	   is true, the data is appended to an existing file
	**/
	bool exportToMatFile(const char* acPath, const char* acVarname, bool bAppend);

	/**
	   Sets the elements of the vector to zero
	**/
	//  void zero(){ for(int i=0;i<iDim;i++) ptData[i]=static_cast<Type>(0);};

	/**
	   Makes a dump of the vector to Sets the elements of the vector to zero
	**/
	void dump(ostream& ostr=cout) const;

	/*!
	 * Return pointer to data array. NO FORGET IT, DON'T USE IT !!
	 **/
	Type* _data() const { return ptData;} ;
	void  _setData(Type *ptNewData);

	bool serialize(fstream& stream, IO_MODE mode = READ);
	bool serialize2(CArchiv& tA);
	bool isSparse() const { return 0; }
	int getNumNonZero() const;
	Type euclideanLength(bool bQuad=false) const;

	Type* begin() { return ptData; }
	Type* end() { return ptData + iDim; }
	Type const* begin() const { return ptData; }
	Type const* end() const { return ptData + iDim; }




}; // end of class

template<class Type>
ostream& operator<<(ostream& stream, const CDenseVector<Type>& vec)
{
	for(const Type* i = vec.begin(); i != vec.end(); ++i) {
		stream << *i << ' ';
	}
	return stream;
}



// CVector operators that will result in a dense vector

// vector-vector ops
template<typename Type>
CDenseVector<Type> operator+(const CVector<Type>& rtVec1, const CVector<Type>& rtVec2)
{
	CDenseVector<Type> rtResult(rtVec1.dimension());
	CVectorOperations<Type>::getInstance().getElementAddition(rtVec1, rtVec2)(rtResult, rtVec1, rtVec2);
	return rtResult;
}

template<class Type> CDenseVector<Type> operator-(const CVector<Type>& rtVec1, const CVector<Type>& rtVec2)
{
	CDenseVector<Type> rtResult(rtVec1.dimension());
	CVectorOperations<Type>::getInstance().getElementSubtraction(rtVec1, rtVec2)(rtResult, rtVec1, rtVec2);
	return rtResult;
}

template<class Type> CDenseVector<Type> operator+(const CVector<Type>& rtVec, Type tScalar)
{
	CDenseVector<Type> rtResult(rtVec);
	rtResult+=tScalar;
	return rtResult;
}
template<class Type> CDenseVector<Type> operator-(const CVector<Type>& rtVec, Type tScalar)
{
	CDenseVector<Type> rtResult(rtVec);
	rtResult-=tScalar;
	return rtResult;
}

// operators w/ by-value returns
template<typename Type>
CDenseVector<Type> elementDiv(const CVector<Type>& v1, const CVector<Type>& v2)
{
	return (CDenseVector<Type>(v1) /= v2);
}

template<typename Type>
CDenseVector<Type> operator*(const CVector<Type>& v, const Type& s)
{
	return (CDenseVector<Type>(v) *= s);
}
template<typename Type>
CDenseVector<Type> operator/(const CVector<Type>& v, const Type& s)
{
	return (CDenseVector<Type>(v) /= s);
}





#endif // CDENSEVECTOR_H
