/*
  COPYRIGHT (C) 2003 APPLIED NEUROINFORMATIC GROUP - UNIVERSITY OF BIELEFELD.

  ALL RIGHTS RESERVED.

  REDISTRIBUTION AND USE IN SOURCE AND BINARY FORM, WITH OR WITHOUT
  MODIFICATION, REQUIRE THE PERMISSION OF THE COPYRIGHT HOLDERS.

  COMMERCIAL USE WITHOUT THE EXPLICIT PERMISSION OF THE COPYRIGHT HOLDERS
  IS FORBIDDEN

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS IS''
  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE LIABLE FOR
  ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/



#ifndef CKERNEL_H
#define CKERNEL_H

#include <cmath>
#include <stdio.h>
#include <assert.h>
#include "CDenseVector.h"
#include "CVector.h"
#include <string>
#include "CObject.h"
#include "CArchiv.h"

/*!
 * @brief Base class for kernel functions
 *  Base class for kernel function used for example by the support vector machine.
 *
 *  @author Thorsten Twellmann
 *  @email ttwellma@TechFak.Uni-Bielefeld.de
 **/
template<class Type>
class CKernel : public CObject<Type>
{
private:

	double m_dBandwidth;

public:
	CKernel(Type tBandwidth=-1.0);
	virtual ~CKernel(){};
	//CKernel<Type>& operator=(const CKernel<Type> rtKernel);

	inline Type bandwidth() const { return m_dBandwidth;};
	void setBandwidth(Type tB) { m_dBandwidth=tB;};
	virtual Type execute(const CVector<Type>& rtVector1, const CVector<Type>& rtVector2) const=0;
	virtual Type execute(const CVector<Type>& rtVector1, const Type* ptVector2) const=0;
	virtual Type execute(const Type* ptVector1, const Type* ptVector2, int iDim) const=0;
	virtual Type execute(Type tValue) const=0;
	virtual CKernel<Type>* clone()const=0 ;
	virtual string className() const { return string("CKernel");};
	virtual bool isA(const char* acClass) const { if(string(acClass) == this->className()) return true; else return CObject<Type>::isA(acClass);};
	virtual string factoryObject()=0;
	virtual bool serialize(fstream& stream, IO_MODE mode );
	virtual bool serialize2(CArchiv& tA);
};


/*!
 * @brief Gaussian kernel function
 *
 * Gaussian kernel function: \f[ K(\mathbf{x},\mathbf{y}) = exp\Bigl(-\frac{\|\mathbf{x}-\mathbf{y}\|^2}{2\sigma^2}\Bigr)  \f]
 *
 **/
template<class Type>
class CGaussianKernel : public CKernel<Type>
{

public:
	CGaussianKernel(Type tBandwidth = 1.0);
	~CGaussianKernel();
	CGaussianKernel(const CGaussianKernel<Type>& tRef);
	CGaussianKernel& operator=(const CGaussianKernel<Type>& rtKernel);
	inline Type execute( const CVector<Type>& rtVector1,  const CVector<Type>& rtVector2) const;
	inline Type execute( const CVector<Type>& rtVector1, const Type* ptVector2)  const;
	inline Type execute( const Type* ptVector1, const Type* ptVector2, int iDim)  const;
	inline Type execute(Type tValue) const;
	virtual string className() const { return string("CGaussianKernel");};
	virtual bool isA(const char* acClass) const { if(string(acClass) == this->className()) return true; else return CKernel<Type>::isA(acClass);};
	string factoryObject() { return string("CGaussianKernelFactory.so");};
	Type gradientBandwidth( const CVector<Type>& rtVector1, const CVector<Type>&  rtVector2, Type tDiagConst = 0);
	CGaussianKernel<Type>* clone() const;




};




/*!
 * @brief Plummer kernel function
 *
 * Plummer kernel function: \f[ K(\mathbf{x},\mathbf{y}) = (\|\mathbf{x}-\mathbf{y}\|_2^2 + \epsilon^2)^{-\frac{\sigma}{2}} \f]
 *
 **/
template<class Type>
class CPlummerKernel : public CKernel<Type>
{
protected:
	double m_dEpsilon;
public:
	CPlummerKernel(Type tBandwidth = 1);
	~CPlummerKernel();
	CPlummerKernel(const CPlummerKernel<Type>& tRef);
	CPlummerKernel& operator=(const CPlummerKernel<Type>& rtKernel);
	inline Type execute( const CVector<Type>& rtVector1,  const CVector<Type>& rtVector2) const;
	inline Type execute( const CVector<Type>& rtVector1, const Type* ptVector2) const ;
	inline Type execute( const Type* ptVector1, const Type* ptVector2, int iDim) const ;
	inline Type execute(Type tValue) const;
	virtual string className() const { return string("CPlummerKernel");};
	virtual bool isA(const char* acClass) const { if(string(acClass) == this->className()) return true; else return CKernel<Type>::isA(acClass);};


	virtual void setEpsilon(Type tE);

	string factoryObject() { return string("CPlummerFactory.so");};
	CPlummerKernel<Type>* clone() const;
	bool serialize(fstream& stream, IO_MODE mode );
	virtual bool serialize2(CArchiv& tA);
};


/*!
 * @brief Triangular kernel function
 *
 * Triangular kernel function: \f[ K(\mathbf{x},\mathbf{y}) = -\frac{\|\mathbf{x}-\mathbf{y}\|^2}{\sigma})  \f]
 *
 **/
template<class Type>
class CTriangularKernel : public CKernel<Type>
{

public:
	CTriangularKernel(Type tBandwidth = 1);
	~CTriangularKernel();
	CTriangularKernel(const CTriangularKernel<Type>& tRef);
	CTriangularKernel& operator=(const CTriangularKernel<Type>& rtKernel);
	inline Type execute( const CVector<Type>& rtVector1,  const CVector<Type>& rtVector2) const;
	inline Type execute( const CVector<Type>& rtVector1, const Type* ptVector2)  const;
	inline Type execute( const Type* ptVector1, const Type* ptVector2, int iDim) const ;
	inline Type execute(Type tValue) const;
	virtual string className() const { return string("CTriangularKernel");};
	virtual bool isA(const char* acClass) const { if(string(acClass) == this->className()) return true; else return CKernel<Type>::isA(acClass);};

	string factoryObject() { return string("CTriangularKernelFactory.so");};
	CTriangularKernel<Type>* clone() const;


};


template<class Type>
class CMultiVariateGaussianKernel : public CKernel<Type>
{

protected:
	CDenseVector<Type> tVecBandwidth;
public:
	CMultiVariateGaussianKernel(CDenseVector<Type> tTmp = CDenseVector<Type>());
	~CMultiVariateGaussianKernel();
	CMultiVariateGaussianKernel(const CMultiVariateGaussianKernel<Type>& tRef);
	CMultiVariateGaussianKernel& operator=(const CMultiVariateGaussianKernel<Type>& rtKernel);
	inline Type execute( const CVector<Type>& rtVector1,  const CVector<Type>& rtVector2) const;
	inline Type execute( const CVector<Type>& rtVector1, const Type* ptVector2)  const;
	inline Type execute( const Type* ptVector1, const Type* ptVector2, int iDim)  const;
	inline Type execute(Type tValue) const;
	void setBandwidth(Type tBw);
	virtual string className() const { return string("CMultiVariateGaussianKernel");};
	virtual bool isA(const char* acClass) const { if(string(acClass) == this->className()) return true; else return CKernel<Type>::isA(acClass);};
	void setBandwidth(CDenseVector<Type>& tBw);

	CMultiVariateGaussianKernel<Type>* clone() const;


	virtual string factoryObject() {return string("");};

};


/*!
 * @brief Polynomial kernel function
 *
 * Polynomial kernel function: \f[ K(\mathbf{x},\mathbf{y}) = \Bigl[\frac{\langle\mathbf{x}\cdot\mathbf{y}\rangle}{\sigma}+r\Bigr]^d \f]
 *
 **/
template<class Type>
class CPolynomialKernel : public CKernel<Type>
{
public:
	double dR;
	int iDegree;

public:
	CPolynomialKernel(Type tBandwidth = 1, int iDegree = 2, Type tR = 0);
	CPolynomialKernel<Type>(const CPolynomialKernel<Type>& rtKernel);
	~CPolynomialKernel();
	CPolynomialKernel<Type>& operator=(const CPolynomialKernel<Type>& rtKernel);
	void setDegree(int iDegree);
	void setR(Type tR);
	virtual string className() const { return string("CPolynomialKernel");};
	virtual bool isA(const char* acClass) const { if(string(acClass) == this->className()) return true; else return CKernel<Type>::isA(acClass);};
	CPolynomialKernel<Type>* clone() const;
	inline Type execute( const CVector<Type>& rtVector1,  const CVector<Type>& rtVector2) const;
	inline Type execute( const CVector<Type>& rtVector1, const Type* ptVector2) const;
	inline Type execute( const Type* ptVector1, const Type* ptVector2, int iDim) const;
	inline Type execute(Type tValue) const;
	string factoryObject() { return string("CPolynomialKernelFactory.so");};
	bool serialize(fstream& stream, IO_MODE mode );
	virtual bool serialize2(CArchiv& tA);
};




/*!
 * @brief  Linear kernel function
 *
 * Linear kernel function: \f[ K(\mathbf{x},\mathbf{y}) = \langle\mathbf{x}\cdot\mathbf{y}\rangle \f]
 *
 **/
template<class Type>
class CLinearKernel : public CKernel<Type>
{
public:
	CLinearKernel();
	CLinearKernel<Type>(const CLinearKernel<Type>& rtKernel);
	CLinearKernel<Type>& operator=(const CLinearKernel<Type>& rtKernel);
	~CLinearKernel();
	virtual string className() const { return string("CLinearKernel");};
	virtual bool isA(const char* acClass) const { if(string(acClass) == this->className()) return true; else return CKernel<Type>::isA(acClass);};
	CLinearKernel<Type>* clone() const;
	inline Type execute(const CVector<Type>& rtVector1, const CVector<Type>& rtVector2) const;
	inline Type execute( const CVector<Type>& rtVector1, const Type* ptVector2) const;
	inline Type execute( const Type* ptVector1, const Type* ptVector2, int iDim) const;
	inline Type execute(Type tValue) const;
	string factoryObject() { return string("CLinearKernelFactory.so");};

};





/*!
 *  @brief Double-Exponential kernel function
 *
 *  Double-exponential kernel function: \f[ K(\mathbf{x},\mathbf{y}) = exp[ -\frac{|\mathbf{x}-\mathbf{y}|}{\sigma}] \f]
 *
 *  @author Thorsten Twellmann
 *  @email ttwellma@TechFak.Uni-Bielefeld.de
 **/
template<class Type>
class CDoubleExponentialKernel : public CKernel<Type>
{

public:
	CDoubleExponentialKernel(Type tBandwidth = 1) ;
	CDoubleExponentialKernel<Type>(const CDoubleExponentialKernel<Type>& rtKernel);
	~CDoubleExponentialKernel();
	CDoubleExponentialKernel<Type>& operator=(const CDoubleExponentialKernel<Type>& rtKernel);
	CDoubleExponentialKernel<Type>* clone() const;
	inline Type execute( const CVector<Type>& rtVector1,  const CVector<Type>& rtVector2) const;
	inline Type execute( const CVector<Type>& rtVector1, const Type* ptVector2) const;
	inline Type execute(  const Type* ptVector1,  const Type* ptVector2, int iDim) const;
	inline Type execute(Type tValue) const;
	string factoryObject() { return string("CDoubleExponentialKernelFactory.so");};
	virtual string className() const { return string("CDoubleExponentialKernel");};
	virtual bool isA(const char* acClass) const { if(string(acClass) == this->className()) return true; else return CKernel<Type>::isA(acClass);};

};


/*!
 *  @brief ANOVA kernel function
 *
 *  Kernel function \f[ K(\mathbf{x},\mathbf{y}) = \Bigl[ \sum_{k=1}^n exp(-\frac{(x_k - y_k)^2}{\sigma}) \Bigr]^d   \f]
 *
 *  @author Thorsten Twellmann
 *  @email ttwellma@TechFak.Uni-Bielefeld.de
 **/
template<class Type>
class CAnovaKernel : public CKernel<Type>
{
protected:
	int iDegree;

public:
	CAnovaKernel(Type tBandwidth=1,int iDegree = 2);
	~CAnovaKernel();
	CAnovaKernel<Type>& operator=(const CAnovaKernel<Type>& rtKernel);
	CAnovaKernel<Type>(const CAnovaKernel<Type>& rtKernel);
	CAnovaKernel<Type>* clone() const ;
	inline void setDegree(int iDegree);
	inline Type execute( const CVector<Type>& rtVector1,  const CVector<Type>& rtVector2) const;
	inline Type execute( const CVector<Type>& rtVector1,  const Type* ptVector2) const;
	inline Type execute(  const Type* ptVector1,  const Type* ptVector2, int iDim) const;
	inline Type execute(Type tValue) const;
	virtual string className() const { return string("CAnoveKernel");};
	virtual bool isA(const char* acClass) const { if(string(acClass) == this->className()) return true; else return CKernel<Type>::isA(acClass);};
	string factoryObject() { return string("CAnovaKernelFactory.so");};
	bool serialize(fstream& stream, IO_MODE mode );
	virtual bool serialize2(CArchiv& tA);
};



/**
   sym approx. of Chi^2  kernel.

   @author Thorsten Twellmann
   @email ttwellma@TechFak.Uni-Bielefeld.de
**/
template<class Type>
class CChiSquaredKernel : public CKernel<Type>
{

public:
	CChiSquaredKernel<Type>(Type tBandwidth = 1);
	~CChiSquaredKernel<Type>();
	CChiSquaredKernel<Type>& operator=(const CChiSquaredKernel<Type>& rtKernel);
	CChiSquaredKernel<Type>(const CChiSquaredKernel<Type>& rtKernel) ;
	CChiSquaredKernel<Type>* clone() const;
	inline Type execute( const CVector<Type>& rtVector1,  const CVector<Type>& rtVector2) const;
	inline Type execute( const CVector<Type>& rtVector1,  const Type* ptVector2) const;
	inline Type execute(  const Type* ptVector1,  const Type* ptVector2, int iDim) const;
	inline Type execute(Type tValue) const;

	virtual string className() const { return string("CChiSquaredKernelFactoryKernel");};
	virtual bool isA(const char* acClass) const { if(string(acClass) == this->className()) return true; else return CKernel<Type>::isA(acClass);};
	string factoryObject() { return string("CChiSquaredKernelFactory.so");};

};

/**
   Epanechnikov
   @author Thorsten Twellmann
   @email ttwellma@TechFak.Uni-Bielefeld.de
**/
template<class Type>
class CEpanechnikovKernel : public CKernel<Type>
{

public:
	CEpanechnikovKernel<Type>(Type tBandwidth = 1);
	~CEpanechnikovKernel<Type>();
	CEpanechnikovKernel<Type>& operator=(const CEpanechnikovKernel<Type>& rtKernel);
	CEpanechnikovKernel<Type>(const CEpanechnikovKernel<Type>& rtKernel) ;
	CEpanechnikovKernel<Type>* clone() const;
	inline Type execute( const CVector<Type>& rtVector1,  const CVector<Type>& rtVector2) const;
	inline Type execute( const CVector<Type>& rtVector1,  const Type* ptVector2) const;
	inline Type execute(  const Type* ptVector1,  const Type* ptVector2, int iDim) const;
	inline Type execute(Type tValue) const;
	string factoryObject() { return string("CEpanechnikovKernelFactory.so");};

};

extern "C"{
	typedef CKernel<double>* _create_double_kernel();
	typedef void _destroy_double_kernel(CKernel<double> *);
	typedef CKernel<float>* _create_float_kernel();
	typedef void _destroy_float_kernel(CKernel<float> *);
}
#endif
