/*
  COPYRIGHT (C) 2003 APPLIED NEUROINFORMATIC GROUP - UNIVERSITY OF BIELEFELD.

  ALL RIGHTS RESERVED.

  REDISTRIBUTION AND USE IN SOURCE AND BINARY FORM, WITH OR WITHOUT
  MODIFICATION, REQUIRE THE PERMISSION OF THE COPYRIGHT HOLDERS.

  COMMERCIAL USE WITHOUT THE EXPLICIT PERMISSION OF THE COPYRIGHT HOLDERS
  IS FORBIDDEN

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS IS''
  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE LIABLE FOR
  ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/



#ifndef CLEARNRATE_H
#define CLEARNRATE_H

#include <cmath>
#include <cstdio>
#include <cassert>
#include "CObject.h"
/******************************************************************************/
/**
   @brief Base class for all learning rate functions
   Template base class for all learning rates (virtual).
**/
template<class Type>
class CLearnRate : public CObject<Type> {

protected:

	Type tInitial;
	Type tFinal;
	Type tRate;
	Type tDecay;
	int  iMaxIteration;
	int  iIteration;

public:

/*----------------------------------------------------------------------------*/
	/**
	   @
	   Gets the learning rate for the next iteration.
	**/
	virtual Type iterate(void) { return 0; }
	virtual ~CLearnRate(){};
/*----------------------------------------------------------------------------*/
	/**
	   Gets the learning rate for a given iteration number.
	**/
	virtual Type rate(int iIterationNumber) { return 0; }

	void setInitial(Type tVal)   { tInitial = tVal; }
	void setFinal(Type tVal)     { tFinal = tVal; }
	void setIterations(int iVal) { iMaxIteration = iVal; }
	void reset()                 { iIteration = 0; tRate = tInitial;}
	void setIteration(int iVal)  { iIteration = iVal; }

	Type getInitial()    { return tInitial; }
	Type getFinal()      { return tFinal; }
	Type getRate()       { return tRate; }
	int  getIterations() { return iMaxIteration; }
	int  getIteration()  { return iIteration; }
	virtual string className() const { return string("CLearnRate"); };
	virtual bool isA(const char* acClass) const {
		if(string(acClass) == this->className())
			return true;
		else
			return  CObject<Type>::isA(acClass);
	};
	virtual bool serialize (fstream &stream, IO_MODE mode=READ);
};

/******************************************************************************/
/**
   @brief Constante learn rate
   Template class for constant learn rates.
**/
template<class Type>
class CConstantRate : public CLearnRate<Type>{

public:

/*----------------------------------------------------------------------------*/
	/**
	   Constructor for a constant learning rate object.
	   @param tRate Defines the learning rate
	**/
	CConstantRate(Type tLearnRate=1.0);

/*----------------------------------------------------------------------------*/
	/**
	   Gets the (constant) learning rate.
	   @return learning rate
	**/

	Type iterate(void);

/*----------------------------------------------------------------------------*/
	/**
	   Gets the (constant) learning rate.
	   @param iIteration Defines the iteration number.
	   @return learning rate
	**/

	Type rate(int iIterationNumber);


	virtual string className() const { return "CConstantRate"; };
	virtual bool isA(const char* acClass) const {
		if(string(acClass) == this->className())
			return true;
		else
			return  CLearnRate<Type>::isA(acClass);
	};
};

/******************************************************************************/
/**
   Template class for harmonic decreasing learning rates.
   The learning rate is defined as initialValue/iterationNumber.
**/
template<class Type>
class CHarmonicRate : public CLearnRate<Type>{

public:

/*----------------------------------------------------------------------------*/
	/**
	   Constructor for a harmonic learning rate object.
	   @param tInitial Defines the initial learning rate
	**/
	CHarmonicRate(Type tInitialRate=1.0);

/*----------------------------------------------------------------------------*/
	/**
	   Gets the learning rate for the next iteration.
	   @return learning rate
	**/

	Type iterate(void);

/*----------------------------------------------------------------------------*/
	/**
	   Gets the learning rate for iteration iIteration.
	   @param iIteration Defines the iteration number.
	   @return learning rate
	**/

	Type rate(int iIterationNumber);

	virtual string className() const { return "CHarmonicRate"; };
	virtual bool isA(const char* acClass) const {
		if(string(acClass) == this->className())
			return true;
		else
			return  CLearnRate<Type>::isA(acClass);
	};
};


/******************************************************************************/
/**
   @brief Linear decreasing learn rate
   Template class for linear decreasing learning rates.
**/
template<class Type>
class CLinearRate : public CLearnRate<Type>{

public:

/*----------------------------------------------------------------------------*/
	/**
	   Constructor for a linear learning rate object.
	   @param tInitialRate Defines the initial learning rate
	   @param tFinalRate Defines the final learning rate
	   @param iMaxIterationNumber Defines the maximum number of iterations
	**/
	CLinearRate(Type tInitialRate=0.5, Type tFinalRate=0.01, int iMaxIterationNumber=100);

/*----------------------------------------------------------------------------*/
	/**
	   Gets the learning rate for the next iteration.
	   @return learning rate
	**/

	Type iterate(void);

/*----------------------------------------------------------------------------*/
	/**
	   Gets the learning rate for iteration iIteration.
	   @param iIteration Defines the iteration number.
	   @return learning rate
	**/

	Type rate(int iIterationNumber);

	virtual string className() const { return "CLinearRate"; };
	virtual bool isA(const char* acClass) const {
		if(string(acClass) == this->className())
			return true;
		else
			return  CLearnRate<Type>::isA(acClass);
	};
};


/******************************************************************************/
/**
   @brief Exponential decreasing learn rate
   Template class for exponential decreasing learning rates.
   The learning rate is defined as initial*(final/initial)^(t/tmax).)
**/
template<class Type>
class CExponentialRate : public CLearnRate<Type>{

public:

/*----------------------------------------------------------------------------*/
	/**
	   Constructor for an exponential learning rate object.
	   @param tInitialRate Defines the initial learning rate
	   @param tFinalRate Defines the final learning rate
	   @param iMaxIterationNumber Defines the maximum number of iterations
	**/
	CExponentialRate(Type tInitialRate=0.5, Type tFinalRate=0.01, int iMaxIterationNumber=100);

/*----------------------------------------------------------------------------*/
	/**
	   Gets the learning rate for the next iteration.
	   @return learning rate
	**/

	Type iterate(void);

/*----------------------------------------------------------------------------*/
	/**
	   Gets the learning rate for iteration iIteration.
	   @param iIteration Defines the iteration number.
	   @return learning rate
	**/

	Type rate(int iIterationNumber);

	virtual string  className() const { return "CExponentialRate"; };
	virtual bool isA(const char* acClass) const {
		if(string(acClass) == this->className())
			return true;
		else
			return  CLearnRate<Type>::isA(acClass);
	};
};

#endif
