/*
  COPYRIGHT (C) 2003 APPLIED NEUROINFORMATIC GROUP - UNIVERSITY OF BIELEFELD.

  ALL RIGHTS RESERVED.

  REDISTRIBUTION AND USE IN SOURCE AND BINARY FORM, WITH OR WITHOUT
  MODIFICATION, REQUIRE THE PERMISSION OF THE COPYRIGHT HOLDERS.

  COMMERCIAL USE WITHOUT THE EXPLICIT PERMISSION OF THE COPYRIGHT HOLDERS
  IS FORBIDDEN

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS IS''
  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE LIABLE FOR
  ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/



#ifndef CMATRIX_H
#define CMATRIX_H

#include <iostream>
#include <fstream>
#include "CObject.h"
using std::ostream;


enum TMatrixType { SYMETRIC };


template<class Type> class CMatrix;
template<class Type> class CVector;
template<class Type> class CDenseVector;





/*- Friends --- Matrix-Matrix-Functions --------------------------------------------------*/

/*!
 *  Elementwise addition of two matrices with returned result.
 **/
template<class Type> CMatrix<Type> operator+(const CMatrix<Type>& rtMat1, const CMatrix<Type>& rtMat2);

/*!
 *  Elementwise subtraction of two matrices with returned result.
 **/
template<class Type> CMatrix<Type> operator-(const CMatrix<Type>& rtMat1, const CMatrix<Type>& rtMat2);

/*!
 *  Elementwise multiplication of two matrices with returned result.
 **/
template<class Type> CMatrix<Type> elementMul(const CMatrix<Type>& rtMat1, const CMatrix<Type>& rtMat2);

/*!
 *  Elementwise division of two matrices with returned result.
 **/
template<class Type> CMatrix<Type> elementDiv(const CMatrix<Type>& rtMat1, const CMatrix<Type>& rtMat2);

/*!
 *  Multiplication of two matrices with returned result.
 **/
template<class Type> CMatrix<Type> operator*(const CMatrix<Type>& tMat1, const CMatrix<Type>& tMat2);


/*- Friends --- Matrix-Scalar-Functions --------------------------------------------------*/

/*!
 *  Addition of a scalar to each element with returned result.
 *^*/
template<class Type> CMatrix<Type> operator+(const CMatrix<Type>& rtMat, const Type& rtScalar);

/*!
 *  Multiplication of a scalar with each element with returned result.
 **/
template<class Type> CMatrix<Type> operator*(const CMatrix<Type>& tMat, const Type& rtScalar);

/*!
 *  Subtraction of a scalar from each element with returned result.
 **/
template<class Type> CMatrix<Type> operator-(const CMatrix<Type>& rtMat, const Type& rtScalar);

/*!
 *  Division of each element by a scalar with returned result.
 **/
template<class Type> CMatrix<Type> operator/(const CMatrix<Type>& rtMat, const Type& rtScalar);


/*- Friends --- Matrix-Vector-Functions --------------------------------------------------*/


/*!
 *  Vector matrix multiplication
 **/
template<class Type> CDenseVector<Type> operator*(const CMatrix<Type>& rtMat, const CVector<Type>& rtVec);


/*!
 *  Save matrix
 **/
template<class Type> void save(const CMatrix<Type>& rtMat, char* acString);


/*!
 *  Load matrix
 **/
template<class Type> void load(CMatrix<Type>& rtMat,char* acString);

/*-Class definition------------------------------------------------------------*/

/*!
 * @brief Template object implementing a matrix of single and double precision elements.

 The elements are stored in a C array in column-major order for easy utilization of LAPACK or CBLAS
 packages.
**/

template<class Type>
class CMatrix : public CObject<Type>
{
	friend CMatrix<Type> operator+<>(const CMatrix<Type>& rtMat1, const CMatrix<Type>& rtMat2);
	friend CMatrix<Type> operator-<>(const CMatrix<Type>& rtMat1, const CMatrix<Type>& rtMat2);
	friend CMatrix<Type> elementMul<>(const CMatrix<Type>& rtMat1, const CMatrix<Type>& rtMat2);
	friend CMatrix<Type> elementDiv<>(const CMatrix<Type>& rtMat1, const CMatrix<Type>& rtMat2);
	friend CMatrix<Type> operator*<>(const CMatrix<Type>& tMat1, const CMatrix<Type>& tMat2);



	friend CMatrix<Type> operator+<>(const CMatrix<Type>& rtMat, const Type& rtScalar);
	friend CMatrix<Type> operator-<>(const CMatrix<Type>& rtMat, const Type& rtScalar);
	friend CMatrix<Type> operator*<>(const CMatrix<Type>& rtMat, const Type& rtScalar);
	friend CMatrix<Type> operator/<>(const CMatrix<Type>& rtMat, const Type& rtScalar);

	friend CDenseVector<Type> operator*<>(const CMatrix<Type>& rtMat, const CVector<Type>& rtVec);

	friend void load<>(CMatrix<Type>&, char* acString);
	friend void save<>(const CMatrix<Type>&, char* acString);
protected:
	int iRows;     /**< Number of rows */
	int iCols;     /**< Number of columns */
	Type* ptData;  /**< Data storage / column-major order */

public:
	/*!
	 *  Constructor -- Create a new instance of a matrix object with
	 *  given size.
	 *
	 *  @param iRows Number of rows (default 0).
	 *  @param iColumns Number of columns (default 0).
	 *  @param bInitialze If true, matrix will be initialized with 0 (default: false).
	 **/
	CMatrix(int iRows = 0, int iColumns = 0, bool bInitialize = false);

	/*!
	 *  Constructor -- Create a new instance of a matrix object which is a
	 *  clone of a second matrix.
	 *
	 *  @param Reference to matrix which will be cloned
	 **/
	CMatrix(const CMatrix<Type>& rtMatrix);

	/*!
	 *  Destructor
	 **/
	~CMatrix();

	/*!
	 *  Assign matrix instance to a matrix variable.
	 *
	 *  @param Matrix.
	 **/
	CMatrix<Type>& operator=(const CMatrix<Type>& rtSrc);

	/*!
	 *  Resize matrix.
	 *  @param iRows Number of rows.
	 *  @param iCols Number of columns.
	 *  @param If true, current matrix data is preserved
	 **/
	void resize(int iRows, int iCols, bool bPreserveData=false);

	/*!
	 *  Return the number of rows of the matrix.
	 *
	 *  @return Number of rows.
	 **/
	inline int rows() const { return iRows; };

	virtual string className() const { return string("CMatrix");};
	/*!
	 *  Return the number of columns of the matrix.
	 *
	 *  @return Number of columns.
	 **/
	inline int columns() const { return iCols; };

	/*!
	 *  Return the value of the element in row i column j.
	 *
	 *  @param i Row index.
	 *  @param j Column index,
	 *  @return Value of element.
	 **/
	const Type& getElement(int i, int j) const;

	/*!
	 *  Return a reference to the element in row i column j.
	 *
	 *  @param i Row index.
	 *  @param j Column index,
	 *  @return Value of element.
	 **/
	Type& operator()(int i, int j);



	/*!
	 * Return the submatrix starting at  row i column j with extend rn and cn.
	 *
	 * @param i Row offset.
	 * @param j Column offset.
	 * @param rn Row extend
	 * @param cn Column extend
	 * @return Submatrix.
	 **/
	CMatrix<Type> operator()(int i, int j, int rn, int cn) const;

	/*!
	 * Insert the submatrix starting at row i column j.
	 *
	 * @param i Row offset.
	 * @param j Column offset.
	 * @param rtSubmatrix.
	 **/
	void insert(int i, int j, const CMatrix<Type> &rtSubmatrix);

	/*!
	 *  Assign a new value to the element in row i column j.
	 *
	 *  @param i Row index.
	 *  @param j Column index,
	 *  @param tValue New value of element.
	 **/
	void setElement(int i, int j, const Type& tValue);

	/*!
	 *  Calculate the determinant of the matrix
	 *
	 *  @return Matrix determinant
	 **/
	Type determinant() const;

	/*!
	 *  Calculate the trace of the matrix.
	 *
	 *  @return Trace of matrix.
	 **/
	Type trace() const;


	/*!
	 *  Sum elements
	 *
	 *  @return  Sum of elements.
	 **/
	Type sumElements() const;

	/*!
	 *  Return a row as a CDenseVector instance.
	 *
	 *  @param iRow Index of row to return.
	 *  @return CDenseVector containing row.
	 **/
	CDenseVector<Type> getRow(int iRow) const;

	/*!
	 *  Return a column as a CDenseVector instance.
	 *
	 *  @param iColumn Index of column to return.
	 *  @return CDenseVector containing column.
	 **/
	CDenseVector<Type> getColumn(int iColumn) const;

	/*!
	 *  Returns the diagonal of the matrix as a CDenseVector instance.
	 *
	 *  @return CDenseVector containing the elements of the matrix diagonal.
	 **/
	CDenseVector<Type> getDiagonal() const;


	/*!
	 * Get index of min element
	 * @param rtX row index of minimal element
	 * @param rtY column index of minimal element
	 * @return Minimal element value
	 **/
	Type minElement(int& rtX, int& rtY) const;


	/*!
	 * Get index of max element
	 * @param rtX row index of maximal element
	 * @param rtY column index of maximal element
	 * @return Maximal element value
	 **/
	Type maxElement(int& rtX, int& rtY) const;
	/*!
	 * Multiply each element by a scalar
	 *
	 *	@param tScalar.
	 **/
	void operator*=(Type rScalar);

	/*!
	 *	 Add scalar to each element
	 *
	 *	 @param tScalar.
	 **/
	void operator+=(Type rScalar);

	/*!
	 *  Subtract from to each element
	 *
	 *  @param tScalar.
	 **/
	void operator-=(Type rScalar);




	/*!
	 *  Divied each element by tScalar
	 *
	 *  @param tScalar.
	 **/
	void operator/=(Type rScalar);

	/*!
	 *  Add matrix
	 *
	 *  @param Matrix.
	 **/
	void operator+=(const CMatrix<Type>& tM);

	/*!
	 *  Subtract matrix
	 *
	 *  @param Matrix.
	 **/
	void operator-=(const CMatrix<Type>& tM);


	/*!
	 *  Fill row iRow with data given by a vector
	 *
	 *  @param iRow Index of row to fill.
	 *  @param Vector containing the new row data.
	 **/
	void setRow(int iRow, const CDenseVector<Type>& tVec);

	/*!
	 *  Fill column iColumn with data given by a vector
	 *
	 *  @param iColumn Index of column to fill.
	 *  @param Vector containing the new column data.
	 **/
	void setColumn(int iIndex, CDenseVector<Type>& tVec);

	/*!
	 *  Set each elements of the diagonal to tValue
	 *
	 *  @param tValue Value.
	 **/
	void  setDiagonal(Type tValue);

	/*!
	 * Fill matrix dagonal with data from vector
	 *
	 * @param Reference to vector with new diagonal elements
	 **/
	void  setDiagonal(CDenseVector<Type>& tDiag);

	/*!
	 *  Calculates the frobenius norm
	 *
	 *  \f[ \parallel A \parallel_F = \sqrt { \sum_i \sum_j a_{ij}^2  } \f]
	 *
	 *  @return Transposed matrix.
	 **/

	Type frobeniusNorm();

	/*!
	 *  Transposes the matrix.
	 *
	 *  @return Transposed matrix.
	 **/
	CMatrix<Type> transpose();

	/*!
	 * Look for NaN elements
	 * @return True if matrix contains NaN elements, false otherwise
	 **/
	bool isNaN() const;

#ifdef UE_ARPACK
	/*!
	 * Matrix-Vector-Product used by ARPACK++
	 */
	void MultMv(Type* v1, Type* v2);
#endif

/**
   Calculates the inverse matrix using LU decomposition.

   @return Inverse matrix
**/
	CMatrix<Type> invert() const;

	/*!
	 * Dump matrix to stream
	 **/
	void dump(ostream& stream=cout);

	bool serialize(fstream& stream, IO_MODE=READ);
	bool serialize2(CArchiv& tArchiv);
/*!
 * Fills matrix with 0
 **/
	void zeros() { for(int i=0; i < iRows*iCols;i++) ptData[i]=0; };

	/*!
	 * Fill matrix with data fromm array.
	 * @param ptData Array of size rows() * cols() with data
	 * @param bColumnMajorOrder If true, data is assumed to be in column-major order, row-major order otherwise (default: false)
	 **/
	void setData(Type* ptData, bool bColumnMajorOrder=false);


	/*!
	 * Export matrix to file MAT-file (Level 1.0)
	 * @param acPath Path to MAType file
	 * @param acVarname Variable name in MAType file
	 * @param bAppend If the file already exists, the data is append to the file (Default: true)
	 * @return True on success, false oterwise
	 **/
	bool exportToMatFile(const char* acPath, const char* acVarame, bool bAppend = true);

	int importFromMatFile(const char* acPath, const char* acVarname);
	/*!
	 * Fill the matrix with a iX x iY mosaic of replicates of the matrix tMat
	 * @param iX X size of mosaic
	 * @param iY Y size of mosaic
	 * @param tMat Single mosaic block
	 **/
	void replicateMatrix(int iX, int iY, const CMatrix<Type> &tMat);

	Type* _data() { return ptData;};



	/*!
	 * Calculate all eigenvalues and eigenvectors of the current matrix. Speifying the type
	 * of the matrix can speed-up the calculation significantly. Additionally, only the first iNum vectors
	 * can be calculated.
	 * @param rtEigenvalues Vector to be filled with eigenvalues in ascending order
	 * @param rtEigenvectors Matrix to be filled with eigenvectors
	 * @param iNumDirections Number of eigenvectors and eigenvalue to calculate
	 * @param tType Matrix type
	 * @return 0 for success, otherwise for error
	 **/
	int eig( CDenseVector<Type>& rtEigenvalues, CMatrix<Type>& rtEigenvectors,  int iNum=0, TMatrixType tType = SYMETRIC);




	/*!
	 *  LU factorization of matrix
	 *
	 **/

	int calcLUFactorization(CMatrix<Type>& rtL, CMatrix<Type>& rtU, CDenseVector<int>& rtIdx) const;




}; // end of class


#endif
