/*
  COPYRIGHT (C) 2003 APPLIED NEUROINFORMATIC GROUP - UNIVERSITY OF BIELEFELD.

  ALL RIGHTS RESERVED.

  REDISTRIBUTION AND USE IN SOURCE AND BINARY FORM, WITH OR WITHOUT
  MODIFICATION, REQUIRE THE PERMISSION OF THE COPYRIGHT HOLDERS.

  COMMERCIAL USE WITHOUT THE EXPLICIT PERMISSION OF THE COPYRIGHT HOLDERS
  IS FORBIDDEN

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS IS''
  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE LIABLE FOR
  ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/



#ifndef CMETRIC_H
#define CMETRIC_H

#include <cmath>
#include <cstdio>
#include <cassert>
#include "CObject.h"
#include "CVector.h"


template<class Type> class CMetric;
template<class Type> class CVector;

/******************************************************************************/
/**
   Template base class for all metrics (virtual).
**/

/*!
 *  @brief Base class for metrix objects
 **/
template<class Type>
class CMetric : public CObject<Type> {
public:
	/**
	   Calculates the distance between rtVec1 and rtVec2 (virtual);
	**/
	virtual Type distance(const CVector<Type>& rtVec1, const CVector<Type>& rtVec2) const { return 0; }
	virtual string className() const { return "CMetric";};
	virtual bool isA(const char* acClass) const { if(string(acClass) == this->className()) return true; else return CObject<Type>::isA(acClass);};
	virtual void adaptVec(CVector<Type>& rtVec1, const CVector<Type>& rtVec2, float fEps) { };
};

/******************************************************************************/
/*!
 * @brief Euclidean metric object
 *
 *  Template class for euclidean metric.
 **/
template<class Type>
class CEuclideanMetric : public CMetric<Type>{
public:

	/**
	   Calculates the euclidean distance between rtVec1 and rtVec2.
	   @param rtVec1 First input Vector.
	   @param rtVec2 Second input Vector.
	   @return euclidean distance.
	**/

	Type distance(const CVector<Type>& rtVec1,const CVector<Type>& rtVec2)  const;
	virtual string className() const { return string("CEuclideanMetric");};
	virtual bool isA(const char* acClass) const { if(string(acClass) == this->className()) return true; else return CMetric<Type>::isA(acClass);};
	void adaptVec(CVector<Type>& rtVec1, const CVector<Type>& rtVec2, float fEps);
};

/******************************************************************************/
/**
 * @brief Euclidean2 metric object
 *
 *Template class for squared euclidean metric.
 **/
template<class Type>
class CEuclideanMetric2 : public CMetric<Type>{
public:

	/**
	   Calculates the squared euclidean distance between rtVec1 and rtVec2.
	   @param rtVec1 First input Vector.
	   @param rtVec2 Second input Vector.
	   @return squared euclidean distance.
	**/

	Type distance(const CVector<Type>& rtVec1, const CVector<Type>& rtVec2)  const;
	virtual string className() const { return string("CEuclideanMetric2");};
	virtual bool isA(const char* acClass) const { if(string(acClass) == this->className()) return true; else return CMetric<Type>::isA(acClass);};
	void adaptVec(CVector<Type>& rtVec1, const CVector<Type>& rtVec2, float fEps);
};

/******************************************************************************/
/**
 * @brief Manhatten metric object
 *
 *Template class for manhatten metric.
 **/
template<class Type>
class CManhattenMetric : public CMetric<Type>{
public:

	/**
	   Calculates the manhatten distance (sum of component differences) between rtVec1 and rtVec2.
	   @param rtVec1 First input Vector.
	   @param rtVec2 Second input Vector.
	   @return manhatten distance.
	**/

	Type distance(const CVector<Type>& rtVec1, const CVector<Type>& rtVec2)  const;
	virtual string className() const { return string("CManhattenMetric");};
	virtual bool isA(const char* acClass) const { if(string(acClass) == this->className()) return true; else return CMetric<Type>::isA(acClass);};
	void adaptVec(CVector<Type>& rtVec1, const CVector<Type>& rtVec2, float fEps);
};

/******************************************************************************/
/**
 * @brief Maximum metric object
 *
 * Template class for maximum metric.
 **/
template<class Type>
class CMaximumMetric : public CMetric<Type>{
public:

	/**
	   Calculates the maximum distance between rtVec1 and rtVec2.
	   @param rtVec1 First input Vector.
	   @param rtVec2 Second input Vector.
	   @return maximum distance.
	**/

	Type distance(const CVector<Type>& rtVec1, const CVector<Type>& rtVec2)  const;
	virtual string className() const { return string("CMaximumMetric");};
	virtual bool isA(const char* acClass) const { if(string(acClass) == this->className()) return true; else return CMetric<Type>::isA(acClass);};
	void adaptVec(CVector<Type>& rtVec1, const CVector<Type>& rtVec2, float fEps);
};

/******************************************************************************/
/**
   @brief Scalar metric object
   Template class for metric based on scalar product.
**/
template<class Type>
class CScalarMetric : public CMetric<Type>{
public:

	/**
	   Calculates the inner product between rtVec1 and rtVec2 -1.0.
	   (resulting in a range from [0.0 .. -2.0] for normalized data)
	   @param rtVec1 First input Vector.
	   @param rtVec2 Second input Vector.
	   @return inner product -1.0.
	**/

	Type distance(const CVector<Type>& rtVec1, const CVector<Type>& rtVec2)  const;
	virtual string className() const { return string("CScalarMetric");};
	virtual bool isA(const char* acClass) const { if(string(acClass) == this->className()) return true; else return CMetric<Type>::isA(acClass);};
	void adaptVec(CVector<Type>& rtVec1, const CVector<Type>& rtVec2, float fEps);
};


/******************************************************************************/
/**
   @brief CoCosine metric object
   Template class for metric based on scalar product.
**/
template<class Type>
class CCoCosineMetric : public CMetric<Type>{
public:

	/**
	   Calculates 1 - scalar product between rtVec1 and rtVec2
	   (resulting in a range from [0.0 .. 2.0] for normalized data)
	   @param rtVec1 First input Vector.
	   @param rtVec2 Second input Vector.
	   @return 1 - scalar product
	**/

	Type distance(const CVector<Type>& rtVec1, const CVector<Type>& rtVec2)  const;
	virtual string className() const { return string("CCoCosineMetric");};
	virtual bool isA(const char* acClass) const { if(string(acClass) == this->className()) return true; else return CMetric<Type>::isA(acClass);};
	void adaptVec(CVector<Type>& rtVec1, const CVector<Type>& rtVec2, float fEps);
};

/******************************************************************************/
/**
 * @brief Angle based metric object
 *
 * Template class for angle based metric.
 **/
template<class Type>
class CAngleMetric : public CMetric<Type>{
public:

	/**
	   Calculates the angle (0-PI) between rtVec1 and rtVec2.
	   @param rtVec1 First input Vector.
	   @param rtVec2 Second input Vector.
	   @return angle.
	**/

	Type distance(const CVector<Type>& rtVec1, const CVector<Type>& rtVec2)  const;
	virtual string className() const { return string("CAngleMetric");};
	virtual bool isA(const char* acClass) const { if(string(acClass) == this->className()) return true; else return CMetric<Type>::isA(acClass);};
	void adaptVec(CVector<Type>& rtVec1, const CVector<Type>& rtVec2, float fEps);
};

/******************************************************************************/
/**
 *  @brief Hyperbolic metric object
 *
 * Template class for hyperbolic metric.
 **/
template<class Type>
class CHyperbolicMetric : public CMetric<Type>{
public:

	/**
	   Calculates the distance of two points in the hyperbolic plane H2
	   @param rtVec1 First input Vector.
	   @param rtVec2 Second input Vector.
	   @return hyperbolic distance.
	**/

	Type distance(const CVector<Type>& rtVec1, const CVector<Type>& rtVec2)  const;
	virtual string className() const { return string("CHyperbolicMetric");};
	virtual bool isA(const char* acClass) const {
		if(string(acClass) == this->className())
			return true;
		else
			return CMetric<Type>::isA(acClass);
	};
	void adaptVec(CVector<Type>& rtVec1, const CVector<Type>& rtVec2, float fEps);
};

#endif
