/*
  COPYRIGHT (C) 2003 APPLIED NEUROINFORMATIC GROUP - UNIVERSITY OF BIELEFELD.

  ALL RIGHTS RESERVED.

  REDISTRIBUTION AND USE IN SOURCE AND BINARY FORM, WITH OR WITHOUT
  MODIFICATION, REQUIRE THE PERMISSION OF THE COPYRIGHT HOLDERS.

  COMMERCIAL USE WITHOUT THE EXPLICIT PERMISSION OF THE COPYRIGHT HOLDERS
  IS FORBIDDEN

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS IS''
  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE LIABLE FOR
  ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/



#ifndef CMULTICLASSSVM_H
#define CMULTICLASSSVM_H

#include "CSVMModel.h"
#include "CKernel.h"
#include "CConjugateGradientOptimizer.h"

template<class Type>
class COneVsAllCalibration : protected CConjugateGradientOptimizer<Type> {
private:
	CDenseVector<Type> m_tTrueClass;
	CMatrix<Type> m_tMargins;
	int m_iClasses;
public:
	COneVsAllCalibration(){};


	Type calibrate(CDenseVector<Type>& tTrueClass, CMatrix<Type>& tMargins) {
		/*-----------------------------------------------------*/
#ifdef SAVE
		ML_EQ_CHK(tTrueClass.dimension(), tMargins.rows());
		ML_GEQ_CHK(tMargins.columns(), m_iClasses);
#endif
/*-----------------------------------------------------*/
		m_tTrueClass = tTrueClass;
		m_tMargins = tMargins;
		m_iClasses = tMargins.columns();
		this->setPosition(CDenseVector<Type>((m_iClasses-1)*2,0.0));
		return optimize();
	};

	CDenseVector<Type> softMax(const CDenseVector<Type>& tVec) const{
		/*-----------------------------------------------------*/
#ifdef SAVE
		ML_EQ_CHK(m_iClasses, tVec.dimension());
#endif
/*-----------------------------------------------------*/

		CDenseVector<Type> tSoftMax(tVec);

		for(int i=0; i<m_iClasses-1; i++)
			tSoftMax[i] = getPosition()[i*2] * tVec[i] + getPosition()[i*2+1];
		tSoftMax[m_iClasses-1] = 1.0 * tVec[m_iClasses-1] + 1.0;
		tSoftMax -= tSoftMax.maxElement();
		for(int i=0; i<m_iClasses; i++)
			tSoftMax[i] = exp(tSoftMax[i]);
		tSoftMax /= tSoftMax.sumElements();
		return tSoftMax;
	};


	bool serialize2(CArchiv& tA) {
		string str;
		int iTmp;
		if(!CConjugateGradientOptimizer<Type>::serialize2(tA))
			return false;

		if(tA.isReading()) {
			tA >> m_iClasses;
			return true;
		} else {
			tA << m_iClasses; tA.flush();
			return true;
		}
		return false;
	};


protected:
	Type functionValueAt(CDenseVector<Type>& tPos) {
		Type tValue=0;

		for(int i=0;i<m_tMargins.rows();i++)
			tValue -= log(softMax(m_tMargins.getRow(i), tPos)[(int)m_tTrueClass[i]] + 0.0000001);
		return tValue;
	};

	CDenseVector<Type> partialDerivativeAt(CDenseVector<Type>& tPos) {
		CDenseVector<Type> tP;
		CDenseVector<Type> tDP(tPos.dimension(),0.0);
		Type tNet;

		for(int i=0;i<m_tMargins.rows();i++) {
			tP = softMax(m_tMargins.getRow(i), tPos);
			for(int j=0; j<m_iClasses-1; j++) {
				if(m_tTrueClass[i]==j)
					tNet = tP[j] - 1.0;
				else
					tNet = tP[j] - 0.0;
				tDP[j*2] += tNet * tP[j] * (1.0-tP[j]) * m_tMargins(i,j);
				tDP[j*2+1] += tNet * tP[j] * (1.0-tP[j]) * 1;
			}
		}
		return tDP;
	};

	CDenseVector<Type> softMax(const CDenseVector<Type>& tVec,const CDenseVector<Type>& tPos) {
		CDenseVector<Type> tSoftMax(tVec);

		for(int i=0; i<m_iClasses-1; i++)
			tSoftMax[i] = tPos[i*2] * tVec[i] + tPos[i*2+1];
		tSoftMax[m_iClasses-1] = 1.0 * tVec[m_iClasses-1] + 1.0;
		tSoftMax -= tSoftMax.maxElement();
		for(int i=0; i<m_iClasses; i++)
			tSoftMax[i] = exp(tSoftMax[i]);
		tSoftMax /= tSoftMax.sumElements();
		return tSoftMax;
	};
};


template<class Type>
class CMultiClassSVM : public CClassifier<Type>
{
public:
	enum TMultiClassSVMType{ ONE_VS_ALL, ONE_VS_ONE };
protected:
	vector< CSVMModel<Type> > m_tClassifiers;
	CDenseVector<int> m_tNumClassItems;
	TMultiClassSVMType m_tType;
	CKernel<Type>* m_ptKernel;
	Type m_tC;
	Type m_tMaxTerminationError;
	bool m_bCalibrated;
	int m_iCacheSize;
	int m_iCalibrationFolds;
	COneVsAllCalibration<Type> m_tOneVsAllCalibration;
public:
	//! Constructor
	CMultiClassSVM(int iNumClasses, TMultiClassSVMType tType = ONE_VS_ALL);


	//! Destructor
	~CMultiClassSVM();


	int classify(const CVector<Type>& tPattern) const;


	void classify(const CVector<Type>&, Type&, Type&) const {};

	int classify(const CVector<Type>&, CDenseVector<Type>& rtVotes) const;
	//! Set kernel functiob
	void setKernel(CKernel<Type>* ptKernel);

	//! Set regularization value
	void setRegularization(Type tC) {this->m_tC = tC;};


	//! Set termination error
	void setMaxTerminationError(Type tTE)  {m_tMaxTerminationError = tTE;};

	void setCalibratedOutput(bool bSwitch) { m_bCalibrated = bSwitch; };


	void trainCalibratedOneVsAll(vector< CDataset<Type> >& rtClassDs);

	//! Set size of kernel cache
	void setCacheSize(int iSize) {m_iCacheSize = iSize;};

	const CKernel<Type>* kernel() const { return m_ptKernel; };

	void train(const CDataset<Type>& rtDataset);
	string className() const;
	bool isA(const char* acClass) const;
	bool serialize2(CArchiv& tA);
protected:
	void trainOneVsAll(vector<CDataset<Type> >& rtClassData);
	void trainOneVsOne(vector<CDataset<Type> >& rtClassData);
};

#endif
