#ifndef CSOM_H
#define CSOM_H

#include "CObject.h"
#include "CDataset.h"
#include "CDenseVector.h"
#include "CDatasetItem.h"
#include "CDatasetStatistic.h"
#include "CMetric.h"
#include "CKernel.h"
#include "CLearnRate.h"
#include "CClusterAlgorithm.h"
#include "Macros.h"

#include "PyramidOfWords.hh"

#include <vector>
#include <set>
#include <map>
#include <ext/hash_set>

#ifdef MYSQL_MODULE
#include "MysqlDatabase.h"
#endif

#define CARTESIAN_SOM       0x01   // plain SOM
#define HYPERBOLIC_SOM      0x02   // hyperbolic SOM
#define H_HYPERBOLIC_SOM    0x04   // hierarchical hyperbolic SOM
#define HG_HYPERBOLIC_SOM   0x08   // hierarchically growing hyperbolic SOM

#define STATS_DIMENSION 11

#define COPY_CLASSCOUNT     0x01
#define COPY_PROTOTYPE      0x02
#define APPEND_BMUS         0x04

////////////////////////////////////////////////////////////////////////////////
/// Base class for different SOM lattice topologies.
/// The base class provides methods to set and get node positions on a
/// regular lattice structure.
////////////////////////////////////////////////////////////////////////////////

using namespace __gnu_cxx;      // we want to use hash_sets

typedef multimap<float, int>    MapBMUs;
typedef map<int, int>           MapPath;
typedef map<int, float>         MapDist;

template<class Type> class CSOM;

class CFlowItem {
public:
	int  *piNodes;
	char acDate[20];
	unsigned long iTimestamp;
};



template<class Type>
class CSomTopology {

	friend class CSOM<Type>;

protected:
	int iType;                            ///< type of lattice topology
	vector< vector<int> > vvNeighbours;   ///< each node has a list of neighbours
	CDataset<double>      tNodes;         ///< dataset of node coordinates
	CDenseVector<double>  tCenter;        ///< vector describing topology center
	vector<int>           viGeometry;     ///< geometry data for VTK

	vector<int>           viVisibleNode;  ///< list of currently visible nodes
	vector<int>           viVisiblePos;   ///< each visible node points to its "real" node

	int                   iCenterNode, iOldCenterNode, iVisibleNodesCenter;
	int                   iNumVisibleNodes;

	vector< pair<int,int> > vDomains;    ///< SOM area might consist of several domains


public:

	vector<bool>          vbExpand;       ///< each node possible carries an expand attribute

	/// Constructor - creates a new instance for given number of nodes
	///
	/// @param iNumNodes number of nodes for new grid
	CSomTopology(int iNumNodes);
	virtual ~CSomTopology();

	/// Sets node position
	/// @param i Number of node to set
	/// @param rtPos Vector describing position of the node on the lattice
	void setNodePosition(int i, CVector<double>& rtPos) { tNodes[i].setInputVector(rtPos); }

	/// Returns node position
	/// @param i Number of node
	/// @return Vector describing position of the node on the lattice
	const CVector<double>& getNodePosition(int i) { return tNodes[i].getInputVector(); }

	const CVector<double>& getCenterPosition() { return tCenter; }

	/// Adds link information between two nodes
	/// @param iFrom Origin node from which the link starts
	/// @param iTo Target node
	void addLink(int iFrom, int iTo);

	/// Returns number of neighbours for specific node
	/// @param i Number of node
	/// @returns Number of neighbours
	int  getNumNeighbours(int i) { return vvNeighbours[i].size(); }

	int  getNumDomains() { return vDomains.size(); }
	void addDomain(int iStart, int iEnd) { vDomains.push_back(make_pair(iStart, iEnd)); }
	int  getDomainStart(int iDomain) { return vDomains[iDomain].first; }
	int  getDomainEnd(int iDomain)   { return vDomains[iDomain].second; }
	int  getDomainSize(int iDomain)  { return vDomains[iDomain].second-vDomains[iDomain].first+1; }
	int  getDomainOfNode(int iNode)  {

		int iDomain;
		for (iDomain=vDomains.size()-1; iDomain > 0; iDomain--)
			if (iNode >= vDomains[iDomain].first) break;
		return iDomain;
	}

	/// Use topology specific training algorithm to build SOM
	virtual void train(CDataset<Type>& rtTrainSet, CSOM<Type>* pSOM) { };
	virtual void trainGrowing(CDataset<Type>& rtTrainSet, CSOM<Type>* pSOM, float fGrowParam) { };

	virtual void buildDomains() { };

	/// Returns node number of specific neighbour of node
	/// @param iNode  Number of node
	/// @param iPos   specifies which neighbour id to return
	/// @returns number of neighbour node
	int  getNeighbour(int iNode, int iPos) { return vvNeighbours[iNode][iPos]; }

	int  getCenterNode(bool bVisibleNodes) { if (bVisibleNodes) return iVisibleNodesCenter;
		else return iCenterNode; }

	/// Returns type of topology
	int  getType()               { return iType; }
	void setType(int iNewType)   { iType = iNewType; }

	/// Returns dimension information
	/// @param viDims Reference to vector. For a two dimensional lattice structure
	/// viDims should contain two vector elements. Each element denotes the
	/// number of nodes along that dimension.
	virtual void getInfo(vector<int>& viDims) { };

	/// Translates grid structure
	virtual int translateGrid(float *pfTrans, float *pfTarget, int iMode=0) { return -1; }

	/// Sets parameters
	virtual void setParameter(float *pfParam) { };

	int     getGeometrySize() { return viGeometry.size(); }
	int     getGeometryPoint(int iPoint) { return viGeometry[iPoint]; }

	virtual int  getNumPolyLines() { return 0; }
	virtual bool getPolyLine(int iLine, int &riStart, int &riEnd) { return false; }

	virtual int getNumOutbounds(int iNode, int iK=0) { };
};


////////////////////////////////////////////////////////////////////////////////
/// Topology class describing two dimensional hyperbolic lattice structure.
////////////////////////////////////////////////////////////////////////////////

template<class Type>
class CHyperbolicTopology : public CSomTopology<Type> {

	friend class CSOM<Type>;

	typedef hash_set<int, hash<int>, equal_to<int> > NodeHashSet;
	typedef set<int, less<int> > NodeSet;

private:
	double fAngle;         ///< Angle in equilateral triangles of hyperbolic tesselation
	double fSin;           ///< Precomputed sinus of triangle angle
	double fCos;           ///< Precomputes cosinus of triangles angle
	double fRadius;        ///< Radius of first ring around center node
	int  iNumNeighbours;   ///< Number of neighbours for each node
	int  iNumRings;        ///< Number of rings for hyperbolic tesselation

	int  iNumVisibleRings;  ///< Number of shown rings
	int  iMaxCenterNode;    ///< Up to which node are iNumVisibleRings present

	double fScale;          ///< scale factor for poincare projection

	NodeHashSet hashsetRingNodes; ///< keeps nodes, which belong to drawn rings (find is fast)
	NodeSet     setRingNodes;     ///< keeps nodes, which belong to drawn rings (sorted)

	int* piRingElements;      ///< is used to compute which nodes belong to shown rings

	vector<int>  viCompleteGeometry; ///< contains visible geometry points
	vector<int>  viOutBounds;        ///< for each node an index points to the vvNeighbour
	///  element, where the outbound region starts

	vector<int>  viPolyLinePoints;
	int          iNumOutbounds;

	set<int, less<int> > setOutbounds;

	int            iBeamWinner;
	int            iBeamBranch;
	bool           bBeamSFsearch;
	int            iBeamLimit;
	bool           bBeamSavePath;

	CVector<Type>  *ptBeamInp;
	CDataset<Type> *pBeamRefs;
	CMetric<Type>  *pBeamMetric;

	MapBMUs        mapBeamArea;
	MapPath        mapBeamPath;
	MapDist        mapBeamDist;

public:
	/// Constructor creates 2 dimensional hyperbolic regular tesselation.
	/// @param iNumNodes Number of nodes in lattice
	/// @param iNeighbours Number of neighbours for each node
	/// @param iRings Number of rings in lattice
	CHyperbolicTopology(int iNumNodes, int iNeighbours, int iRings, int iNumVisible=3);
	virtual ~CHyperbolicTopology();

	/// Computes the coordinates of the lattice structure on Poincare Disk
	void makeLattice();

	/// Rotates node around given center node determinded by angle of
	/// equilateral triangles of tesselation.
	/// @param iCenter Number of center node
	/// @param iFrom Number of origin node
	/// @param iTarget Number of target node
	void rotateNode(int iCenter, int iFrom, int iTarget);

	/// Adds complete ring to lattice structure
	/// @param iFirst Number of first node in new ring
	/// @param iLast Number of last node in inner ring
	int  addRing(int iFirst, int iLast);

	void addTriangle(int i1, int i2, int i3);

	void getInfo(vector<int>& viDims);

	int  translateGrid(float *pfTrans, float *pfTarget, int iMode=0);

	void copyVisibleNodeCoordinates(float *pfTarget);

	void setParameter(float *pfParam);

	void setNumVisibleRings(int iNumVisibleRings);

	int  getNumNodesUptoRing(int iRing);

	/// Returns number of polylines which are use to draw structure
	/// information related to the topology
	int  getNumPolyLines()        { return viPolyLinePoints.size() / 2; }
	/// Returns true if requested polyline iLine is closed
	bool getPolyLine(int iLine, int &riStart, int &riEnd);

	void train(CDataset<Type>& rtTrainSet, CSOM<Type>* pSOM);
	void trainGrowing(CDataset<Type>& rtTrainSet, CSOM<Type>* pSOM, float fGrowParam);

	void performBeamSearch(int iRoot, int iLevel);

	void buildDomains();

	void addPolyLinesFrom(int iNode);

	void computeOutbounds(int iNode);
	void computeOutbounds(int iNode, int iK);

	int  getNumOutbounds(int iNode, int iK=0) {
		iNumOutbounds = 0;
		setOutbounds.clear();
		if (iK) {
			computeOutbounds(iNode, iK);
			return setOutbounds.size();
		} else {
			computeOutbounds(iNode);
		}
		if (iNumOutbounds == 0) return 1;
		else return iNumOutbounds;
	}

	int getOutBounds(int iNode) { return viOutBounds[iNode]; };

	void getOutBoundRange(int iNode, vector<int>::iterator& itStart, vector<int>::iterator& itEnd)  {
		itStart = this->vvNeighbours[iNode].begin() + viOutBounds[iNode];
		itEnd = this->vvNeighbours[iNode].end();
	};

	void setBeamBranching(int iBranch) {
		this->iBeamBranch = abs(iBranch);
		if (iBranch < 0)
			this->bBeamSFsearch = true;
	};
};

////////////////////////////////////////////////////////////////////////////////
/// Topology class for n dimensional cartesian grid
////////////////////////////////////////////////////////////////////////////////
template<class Type>
class CCartesianTopology : public CSomTopology<Type> {

public:
	CCartesianTopology(int iNodes, vector<int>& rviNbUnits);
	virtual ~CCartesianTopology() { }
	vector<int> viDimensions;

	void getInfo(vector<int>& viDims);

	int  translateGrid(float *pfTrans, float *pfTarget, int iMode=0);

	void train(CDataset<Type>& rtTrainSet, CSOM<Type>* pSOM) { pSOM->train(rtTrainSet); }
	void buildDomains();

	int  getNumOutbounds(int iNode, int iK=0) { return 4; }
};


////////////////////////////////////////////////////////////////////////////////
/// SOM class for Self Organizing Maps with arbitrary topology.
////////////////////////////////////////////////////////////////////////////////

template<class Type>
class CSOM : public CClusterAlgorithm<Type> {

#ifdef MYSQL_MODULE
	friend class CMysqlDatabase;
#endif
	friend class CHyperbolicTopology<Type>;

	typedef set<int, less<int> > NodeSet;

private:

	CSomTopology<Type> *ptTopology;   ///< The topology of the map
	CMetric<double>    *ptGridMetric; ///< Metric for computation of node distances on grid
	CMetric<Type>      *ptDataMetric; ///< Metric for computation of distances in dataset

	CLearnRate<Type>   *ptAlphaRate;  ///< Learning rate for learning stepwidth alpha
	CLearnRate<Type>   *ptSigmaRate;  ///< Learning rate for neighborhood width
	CKernel<Type>      *ptKernel;     ///< Kernel to use for neighborhood function

	CDataset<Type>     tMetaData;     ///< Meta data: number of hits per node etc.

	vector<string>     vLabel;        ///< Each node can carry a label
	vector<char*>      vKeyWords;     ///< Each node might carry a list of key words

	vector<vector< pair<float,int> > > vvIDs; ///< Each node can carry a list of IDs

	int iNumNodes;              ///< Number of nodes in SOM
	int iUpdateSteps;           ///< Number of update steps to use for training
	int iBestMatchID;

	void makeReferenceVectors(); ///< allocates data for reference vectors

	NodeSet           setSelectedNodes;
	NodeSet::iterator itNodeSet;

	vector<CFlowItem> vFlowItems;

	vector<CDenseVector<float> >  tActivities;

	float fProgress;

public:
	vector<string>    vTopicNames;    ///< Meta data: topic names

public:
	/// Constrcutor creates a new SOM instance
	/// @param ptTheGridMetric The metric to use for distance computation of nodes
	/// on the grid.
	CSOM(CMetric<double>* ptTheGridMetric = new CEuclideanMetric<double>());

	/// Copy constructor
	CSOM(const CSOM<Type> &SOM);

	/// Descructor
	~CSOM();

	/// Set the grid metric
	/// @param ptMetric The new metric to use.
	void setGridMetric(CMetric<double>* ptMetric);

	/// Set the data metric
	/// @param ptMetric The new metric to use for computation of distances in the
	/// data space - This is used to compute the best match node.
	void setDataMetric(CMetric<Type>* ptMetric);

	CMetric<Type>* getDataMetricPtr() { return ptDataMetric; }

	/// Set the learning rate for the learning stepwidth alpha
	/// @param ptRate The new alpha rate.
	void setAlphaRate(CLearnRate<Type>* ptRate);

	/// Set the rate for the neighborhood width
	/// @param ptRate The new learning rate for the neighborhood parameter sigma.
	void setSigmaRate(CLearnRate<Type>* ptRate);

	/// Set the kernel for the neighborhood function.
	/// @param ptTheKernel The new kernel to use for computation of h.
	void setKernel(CKernel<Type>* ptTheKernel);

	CSomTopology<Type>* topologyPtr() { return ptTopology; }

	/// Returns the meta data dataset containing some statistics like node sizes
	CDataset<Type>& metaData()        { return tMetaData; }
	void setMetaData(int iPos, Type *ptTheData);

	float getMeta4Node(int iNode, int iPos) { return tMetaData[iNode].inputVectorPtr()->getElement(iPos); }

	void setUpdateSteps(int iVal)     { iUpdateSteps = iVal; }

	CLearnRate<Type>& getAlphaRate()  { return *ptAlphaRate; }
	CLearnRate<Type>& getSigmaRate()  { return *ptSigmaRate; }
	CKernel<Type>&    getKernel()     { return *ptKernel; }

	void   setLabel(int iNode, string strLabel);
	string getLabel(int iNode);

	/// Throws an entire dataset onto the SOM. Copy class count statistics or
	/// reference vectors to meta statistics or append list of BMUs for each
	/// input vector of the dataset to the output vector depending on the value
	/// of iMode.
	/// @param iMode Switch between different modes:
	/// COPY_CLASSCOUNT (1), COPY_PROTOTYPE (2), APPEND_BMUS (4)
	void applyDataset(CDataset<Type>& rtDataset, int iMode=COPY_CLASSCOUNT);

	void setUnit(int iNode, CDenseVector<Type>& rtVec);

	/// Initializes the position of the nodes in the output space.
	/// The function set the node vectors according to an equidistant grid, while
	/// the number of nodes in each dimension can be specified using rviNbUnits.
	/// @param rviNbUnits Reference to an vector holding the number of nodes/dimension.
	int setCartesianTopology(vector<int>& rviNbUnits);

	/// Initializes the position of the nodes in the output space.
	/// Places the nodes on a regular grid obtained by a tessellation
	/// of the hyperbolic plane H2 with congruent triangles.
	/// @param iNeighbours Number of neighbours each node will have
	/// @param iRings      Number of rings around center node
	/// @param iVisible    Number of visible rings which are passed to VTK
	int setHyperbolicTopology(int iNeighbours, int iRings, int iVisible=3);

	/// Initializes the position of the prototype vectors in the input space.
	/// The function set prototype vectors according to a given dataset.
	/// @param rtData Reference to an CDataset.
	void setUnitsFromDataset(CDataset<Type>& rtData);

	/// Initializes the position of the prototype vectors in the input space.
	/// The function set prototype vectors according to a random selection of vectors
	/// from a given dataset.
	/// @param rtData Reference to an CDataset.
	void setUnitsFromDatasetRandom(CDataset<Type>& rtData);

	/// Initializes the position of the prototype vectors in the input space.
	/// The function set prototype vectors along the given vectors.
	/// @param rtData Reference to an CDataset.
	void setUnitsAlongVectors(vector<CDenseVector<Type> > cVectors,
	                          vector<Type> cMin,
	                          vector<Type> cMax);


	void train(CDataset<Type>& rtTrainSet);

	void trainTopology(CDataset<Type>& rtTrainSet) {
		if (ptTopology) ptTopology->train(rtTrainSet, this);
	}

	void trainGrowing(CDataset<Type>& rtTrainSet, float fGrowParam) {
		if (ptTopology) ptTopology->trainGrowing(rtTrainSet, this, fGrowParam);
	}

	CDataset<double>& nodeData() {
		return ptTopology->tNodes;
	}

	/// Calculate minimal wiring.
	/// @param ctTestSet  Reference to a test set.
	/// @param ctMetric   Reference to a metric.
	Type minimal_wiring(CMetric<Type>& cMetric);

	/// Dumps the entire SOM to cout.
	void dump(void);

	/// Calculates the average inter-node distance between a reference vector and
	/// it neighbors in a given distance in the output space.
	/// @param rvtDist Reference to a vector with the average inter-node distance.
	/// @param fNeighborhood Range of the neighborhood (default 1.0).
	void uMat(vector<Type>& rvtDist, Type fNeighborhood);

	int  getNumNodes() { return iNumNodes; }

	int  getNumVisibleNodes() {
		if (ptTopology) return ptTopology->iNumVisibleNodes;
		else return 0;
	}

	int  getVisibleNode(int iPos) {
		if (ptTopology) return ptTopology->viVisibleNode[iPos];
		else return -1;
	}



	void copyNodeAttributes(int iContext, int iAttrib, float *pfTarget,
	                        float &fMin, float &fMax, float *pfSource=NULL);

	void copyNodeDistances(float *pfDist);

	bool selectNode(int iNode, bool bSelect);
	int  getNumSelectedNodes() { return setSelectedNodes.size(); }
	int  getSelectedNodes(bool bFirst);

	int  getBestMatchID() { return iBestMatchID; }

	Type getNodeDistance(int iN1, int iN2);

	void setFlowSize(int iNumItems);
	int  getFlowSize()          { return vFlowItems.size(); }

	void setFlowItem(int iPos, int *piNodes, char *pcDate, unsigned long iTime);
	void computeFlow(float *pfParams);
	void getFlow(int iPos, float *pfTarget, char *pcDate);

	// passes all children nodes of node iNode to piChildren. Returns number of children
	int getChildren(int iNode, int* piChildren);

	float getProgress() { return fProgress; }

};

#endif
