/*
  COPYRIGHT (C) 2003 APPLIED NEUROINFORMATIC GROUP - UNIVERSITY OF BIELEFELD.

  ALL RIGHTS RESERVED.

  REDISTRIBUTION AND USE IN SOURCE AND BINARY FORM, WITH OR WITHOUT
  MODIFICATION, REQUIRE THE PERMISSION OF THE COPYRIGHT HOLDERS.

  COMMERCIAL USE WITHOUT THE EXPLICIT PERMISSION OF THE COPYRIGHT HOLDERS
  IS FORBIDDEN

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS IS''
  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE LIABLE FOR
  ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/



#ifndef CSVCANDIDATE_H
#define CSVCANDIDATE_H

#include <assert.h>
#include <Macros.h>
#include <CArchiv.h>
#include "CDatasetItem.h"

/**
   This class implmements datasetitems used for training a
   support vector machine. Addionally to the input and output
   pattern, each item contains a lagrange coefficient and
   an error cache.

   @author Thorsten Twellmann
**/

enum SVTYPE { ACTIVE     = 0,
              BOUNDED    = 1,
              NONSUPPORT = 2
};

enum SORTING{ OUTERROR      = 0,
              LAGRANGE   = 1,
              CLASS      = 2,
              WORKINGSET = 3,
              ID         = 4
};




template<class Type>
class CSVCandidate : public CDatasetItem<Type>
{
public:
	SORTING tSorting;
protected:
	double dLagrange;
	double dGradient;
	Type tTargetValue;
	Type tError;
	double dOutput; /* used for one-class svm */
	int iIdxTestSet;
	int iLastChange;
	SVTYPE tType;


public:









	CSVCandidate() : CDatasetItem<Type>(0,1)
	{
		this->dLagrange   = 0;
		this->tError      = 0;
		this->dGradient   = 0;
		this->iLastChange = 0;
		this->iIdxTestSet = 0;
		this->iId         = 0;
		this->tSorting    = LAGRANGE;
		this->tType       = NONSUPPORT;
		this->tTargetValue= 0;
		this->dOutput     = 0;
	};
	CSVCandidate(int iInputDim) : CDatasetItem<Type>(iInputDim,1)
	{
		this->dLagrange   = 0;
		this->tError      = 0;
		this->dGradient   = 0;
		this->iLastChange = 0;
		this->iIdxTestSet = 0;
		this->iId         = 0;
		this->tSorting    = LAGRANGE;
		this->tType       = NONSUPPORT;
		this->tTargetValue= 0;
		this->dOutput     = 0;
	};

	CSVCandidate(const CVector<Type>& tInput, Type tClass) : CDatasetItem<Type>(tInput,CDenseVector<Type>(1,&tClass))
	{
		this->dLagrange   = 0;
		this->tError      = 0;
		this->dGradient   = 0;
		this->iLastChange = 0;
		this->iIdxTestSet = 0;
		this->iId         = 0;
		this->tSorting    = LAGRANGE;
		this->tType       = NONSUPPORT;
		this->tTargetValue= tClass;
		this->dOutput     = 0;
	};



	CSVCandidate(const CSVCandidate<Type>& rtItem) : CDatasetItem<Type>(rtItem)
	{

		this->tSorting    = rtItem.tSorting;

		this->dLagrange   = rtItem.dLagrange;
		this->tError      = rtItem.tError;
		this->tType       = rtItem.tType;
		this->iId         = rtItem.iId;
		this->dGradient   = rtItem.dGradient;
		this->iIdxTestSet = rtItem.iIdxTestSet;
		this->dOutput   = rtItem.dOutput;
		this->iLastChange = rtItem.iLastChange;
		this->tTargetValue= rtItem.tTargetValue;
		this->tTargetValue= rtItem.tTargetValue;

	};


	CSVCandidate(const CDatasetItem<Type>& rtItem) : CDatasetItem<Type>(rtItem)
	{
		this->tSorting    = LAGRANGE;

		this->dLagrange   = 0;
		this->tError      = 0;
		this->iId         = 0;
		this->tType       = NONSUPPORT;
		this->iIdxTestSet = 0;
		this->dGradient   = 0;
		this->iLastChange = 0;
		this->dOutput     = 0;
		this->tTargetValue= rtItem.getOutputComponent(0);

	};

	CSVCandidate<Type>& operator= (const CSVCandidate<Type>& rtItem)
	{

		if(this == &rtItem) return *this;
		this->tSorting    = rtItem.tSorting;
		this->dLagrange   = rtItem.dLagrange;
		this->tError      = rtItem.tError;
		this->tType       = rtItem.tType;
		this->iId         = rtItem.iId;
		this->dGradient   = rtItem.dGradient;
		this->dOutput   = rtItem.dOutput;
		this->iIdxTestSet = rtItem.iIdxTestSet;
		this->iLastChange = rtItem.iLastChange;
		this->tTargetValue= rtItem.tTargetValue;
		*((CDatasetItem<Type>*) this)=rtItem;

		return *this;
	};

	~CSVCandidate() {};


	bool operator== (const CSVCandidate<Type>& right) const {
		if(tSorting == OUTERROR) {
			if(this->tError == right.tError)
				return true;
			else
				return false;
		} else if(tSorting == LAGRANGE) {
			if(this->dLagrange == right.dLagrange)
				return 1;
			else
				return 0;
		} else if(tSorting == CLASS) {
			if(tTargetValue == right.tTargetValue)
				return 1;
			else
				return 0;
		} else if(tSorting == ID) {
			if(this->iId == right.iId)
				return 1;
			else
				return 0;
		}
		return 0;
	};

	virtual string className() const { return string("CSVCandidate");};
	virtual bool isA(const char* acClass) const
	{
		if(string(acClass) == this->className())
			return true;
		else return CDatasetItem<Type>::isA(acClass);
	};

	bool operator!=(const CSVCandidate<Type>& right) const {
		if(tSorting == CLASS) {
			if(this->tTargetValue != right.tTargetValue)
				return 1;
			else
				return 0;
		} else if(tSorting == LAGRANGE) {
			if(this->dLagrange !=right.dLagrange)
				return 1;
			else
				return 0;
		} else if(tSorting == OUTERROR) {
			if(this->tError != right.tError)
				return 1;
			else
				return 0;
		} else if(tSorting == ID) {
			if(this->iId != right.iId)
				return 1;
			else
				return 0;
		}
		return 0;
	}

	bool operator>(const CSVCandidate<Type>& right) const {
		if(tSorting == CLASS) {
			if(this->tTargetValue < right.tTargetValue)
				return 1;
			else
				return 0;
		} else if(tSorting == LAGRANGE) {
			if(this->dLagrange < right.dLagrange)
				return 1;
			else
				return 0;
		} else if(tSorting == OUTERROR) {
			if(this->tError < right.tError)
				return 1;
			else
				return 0;
		} else if(tSorting == ID) {
			if((int)this->iId < (int)right.iId)
				return 0;
			else
				return 1;
		}
		return 0;
	}

	bool operator<(const CSVCandidate<Type>& right) const {
		if(tSorting == CLASS) {
			if(this->tTargetValue > right.tTargetValue)
				return 1;
			else
				return 0;
		} else if(tSorting == LAGRANGE) {
			if(this->dLagrange > right.dLagrange)
				return 1;
			else
				return 0;
		} else if(tSorting == OUTERROR) {
			if(this->tError > right.tError)
				return 1;
			else
				return 0;
		} else if(tSorting == ID) {
			if((int)this->iId > (int)right.iId)
				return 0;
			else
				return 1;
		}
		return 0;
	}

	inline double& lagrange() {
		return dLagrange;
	};

	inline Type& error() {
		return tError;
	};

	inline int testSetIndex() {
		return iIdxTestSet;
	};

	inline SVTYPE& type() {
		return tType;
	};

	inline Type& targetValue() {
		return tTargetValue;
	};

	inline int& lastChange() {
		return iLastChange;
	};

	inline double& gradient() {
		return dGradient;
	};



	inline double& output() {
		return dOutput;
	};

	inline int& id() {
		return this->iId;
	};




	inline  const SVTYPE& getType()  const {
		return tType;
	};

	inline  const Type& getTargetValue()  const {
		return tTargetValue;
	};

	inline  const double& getLagrange()  const {
		return dLagrange;
	};

	inline  const int& getLastChange()  const {
		return iLastChange;
	};

	inline  const double& getGradient()  const {
		return dGradient;
	};



	inline  const double& getOutput()  const {
		return dOutput;
	};

	inline  const int& getId()  const {
		return this->iId;
	};


	bool serialize(fstream& stream, IO_MODE mode) {
		CDatasetItem<Type>::serialize(stream, mode);
		if(mode == WRITE) {
			int iType=this->dataType();
			stream.write((char*) &iType,        sizeof(int));
			stream.write((char*) &dLagrange,    sizeof(double));
			stream.write((char*) &dGradient,    sizeof(double));
			stream.write((char*) &tTargetValue, sizeof(Type));
			stream.write((char*) &tError,       sizeof(Type));
			stream.write((char*) &dOutput,      sizeof(double));
			stream.write((char*) &iIdxTestSet,  sizeof(int));
			stream.write((char*) &iLastChange,  sizeof(int));
			stream.write((char*) &tType,        sizeof(int));

			return true;
		} else {
			DATATYPE tType;
			stream.read((char*) &tType,        sizeof(int));
			stream.read((char*) &dLagrange,    sizeof(double));
			stream.read((char*) &dGradient,    sizeof(double));
			readValueOfType(tTargetValue, stream, tType);
			readValueOfType(tError, stream, tType);
			stream.read((char*) &dOutput,      sizeof(double));
			stream.read((char*) &iIdxTestSet,  sizeof(int));
			stream.read((char*) &iLastChange,  sizeof(int));
			stream.read((char*) &tType,        sizeof(int));

			return true;
		}
		return false;
	}




	bool serialize2(CArchiv& tA) {
		if(!CDatasetItem<Type>::serialize2(tA))
			return false;

		if( tA.isReading()) {
			int iTmp;
			tA >> dLagrange;
			tA >> dGradient;
			tA >> tTargetValue;
			tA >> tError;
			tA >> dOutput;
			tA >> iIdxTestSet;
			tA >> iLastChange;
			tA >> iTmp; tType = (SVTYPE) iTmp;
			tA >> iTmp; tSorting = (SORTING) iTmp;
			return true;
		} else {
			tA << dLagrange; tA.flush();
			tA << dGradient; tA.flush();
			tA << tTargetValue; tA.flush();
			tA << tError; tA.flush();
			tA << dOutput; tA.flush();
			tA << iIdxTestSet; tA.flush();
			tA << iLastChange; tA.flush();
			tA << (int)tType; tA.flush();
			tA << (int)tSorting; tA.flush();
			return true;
		}
		return false;
	}
};



#endif
