/*
  COPYRIGHT (C) 2003 APPLIED NEUROINFORMATIC GROUP - UNIVERSITY OF BIELEFELD.

  ALL RIGHTS RESERVED.

  REDISTRIBUTION AND USE IN SOURCE AND BINARY FORM, WITH OR WITHOUT
  MODIFICATION, REQUIRE THE PERMISSION OF THE COPYRIGHT HOLDERS.

  COMMERCIAL USE WITHOUT THE EXPLICIT PERMISSION OF THE COPYRIGHT HOLDERS
  IS FORBIDDEN

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS IS''
  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE LIABLE FOR
  ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/



#ifndef CSVM_H
#define CSVM_H

#include "CDenseVector.h"
#include "CDataset.h"
#include "CClassifier.h"
#include "CSVCandidate.h"
#include "CKernelCache.h"

#include "CKernel.h"
#include <vector>

#define SVEXCP_BAD_LABEL "Illegal label detected! Label must be '+1' or '-1'."
#define SVEXCP_BAD_DATASET "Number of classes differs from two!"

#define WARNING_THRESHOLD 1e-11

enum TSVMType {
	L1SVM,
	L2SVM
};

/*!
 * @brief Support Vector Machine algorithm for binary classifcation.
 *
 * Solves the binary SVM training by decomposition. This implementation allows L1 and L2 penalties,
 * class specific regularization, kernel caching and shrinking. The implementation is strongly inspired by \em libsvm but is much more readable. You can
 * find more information about the implementation details in papers about \em Sequential \em Minimal \em Optimization \em
 * and \em libsvm.
 *
 *
 * @author Thorsten Twellmann (ttwellma@TechFak.Uni-Bielefeld.de)
 **/
template <class Type>
class CSVM : public CClassifier<Type> {



protected:
	vector< CSVCandidate<Type> > tWorkingSet; /**< current workingset */
	CKernelCache<Type> *ptCache;              /**< Cache for the kernelmatrix */
	CKernel<Type>* ptKernel;                  /**< Kernel object */
	Type tC;                                  /**< Regularization parameter for L1-SVM*/
	Type tWeightPos, tWeightNeg;              /**< */
	Type tTerminationError;                   /**< Error threshold */
	Type tBias;                               /**< Bias value */
	long lMaxIterations;                      /**< Max number of iterations */
	long lNextShrink;
	bool bUnshrinked;
	bool bFoldKernel;                         /**< If true, the linear kernel is folded to a single vector describing the hyperplane */
	int iNumIteration;                        /**< Current number of iterations */
	int iNumActiveItems;                      /**< Current number of active items */
	TSVMType tSVMType;                        /**< Type of error penalty */
	CDenseVector<Type> tWeight;               /**< speed-up for linear kernel */
	int iCacheSize;
	/* some status flags */
	vector<int> tTransformedAccess;

public:
	/*!
	 * Constructor - Creates a new instance of a SVM with given penalty metric
	 *
	 * @param tSVMType Type of svm (default: L1SVM).
	 **/
	CSVM(TSVMType = L1SVM);

	/*!
	 *  Destructor
	 **/
	~CSVM ();

	/*!
	 * Set the maximal number of iterations
	 * @param lMax Maximal number of iterations
	 **/
	void setMaxIterations (long lMax) {
		lMaxIterations = lMax;
	};



	/*!
	 * Classifies a given pattern.
	 *
	 * @param tVec Vector containing the pattern to classify.
	 * @param ptClass Pointer to variable which will be filled with the classlable.
	 * @param ptMargin Pointer to variable which will be filled with calculated margin.
	 **/
	void classify (const CVector < Type > &tVec, Type* ptClass,Type* ptMargin) const;
	void classify (const CVector < Type > &tVec, Type& rtClass,Type& rtMargin) const;
	/*!
	 * Classifies a given pattern.
	 *
	 * @param ptVec Array containing the pattern to classify.
	 * @param ptClass Pointer to variable which will be filled with the classlable.
	 * @param ptMargin Pointer to variable which will be filled with calculated margin.
	 **/
	void classify (Type * ptVec, Type * ptClass, Type * ptMargin) const;

	int classify (const CVector < Type > &tVec) const ;

	string className() const { return "CSVM";};

	bool isA(const char* acClass) const { if(string(acClass) == "CSVM") return true; else return CClassifier<Type>::isA(acClass);};


	void train (const CDataset<Type>& tTrainSet);

	/*!
	 *  Sets regularization paramert C. C will be the upper bound for the lagrange coefficients of
	 *  the svm in case of L1 penalty and an additional term on the kernel in case of L2 penalty.
	 *
	 *  @param tC New regularization value.
	 **/
	void setRegularization (Type tC);

	/*!
	 *  Set class-specific regularization values. In the case of L1 penalty, the lagrange coefficients \f$ \alpha_i\f$  are upper bounded by the regularization
	 * values \f$ C^+ = w^+ C \f$ and \f$ C^- = w^- C \f$. Typical selections are e.g  \f$ \frac{C^+}{C^-}= \frac{n^-}{n+} \f$ (according Morik, Brockhausen and Joachims 1999),
	 * leading to \f$ w^+ = \frac{n^-}{N} \f$ and \f$ w^- = \frac{n^+}{N} \f$. In the case of L2 penalty, the regularization
	 * values \f$ C^+ = \frac{1}{w^+ C} \f$ and \f$ C^- = \frac{1}{w^- C} \f$ are added to the diagonal of the kernel matrix.
	 * @param tWPos Weight \f$ w^+ \f$ for class \f$ y=+1 \f$
	 * @param tWNeg Weight \f$ w^- \f$ for class \f$ y=-1 \f$
	 **/
	void setWeight(Type tWPos, Type tWNeg);


	/*!
	 *  Sets the termination error. SVM training will terminate if the maximal difference
	 *  between two lagrange gradients will drop below this threshold.
	 *
	 *  @param tMinError Termination error.
	 **/
	void setTerminationError (Type tMinError) {
		this->tTerminationError = tMinError;
	};



	void trainChunking(CDataset<Type>& rtTrainset, int iStartSetSize,int iMaxChunkSize);


/*!
 * Use a new kernel object. The SVM becomes the owner of the kernel
 * object and will delete it in destruction.
 *
 * @param ptKernel Pointer to new kernel object
 **/
	void setKernel(CKernel<Type>* ptKernel);



	/*!
	 *
	 * Return const pointer to kernel object
	 *
	 * @return Const pointer to kernel object
	 **/
	const CKernel<Type>* kernel() {
		return ptKernel;
	};


	/*! Set new cache size (in MByte)
	 * @param tSize Cache size
	 **/
	void setCacheSize (Type tSize) {
		iCacheSize = (int) tSize;
	};




/*!
 * Returns the svm solution as a dataset. The input vector of each item contains
 * the training pattern. The output vector contains the class label (element 0) and
 * the lagrange coefficient (element 1). By setting the parameter bDropNonSV=true, training
 * vectors with lagrange coefficient equal to zero will be deleted.
 * @param bDropNonSV If true, non-support vectors will be deleted
 * @return Dataset containing support data
 **/
	CDataset<Type> getSupportData (bool bDropNonSV);
	CDataset<Type> getNonSupportData() const;




	CDenseVector<Type> getWeightVector() const { return tWeight;};


	/*!
	 * Return bias value
	 * @return Bias
	 **/
	Type bias() const { return tBias;};






	bool serialize(fstream& stream, IO_MODE mode);
	bool serialize2(CArchiv& tA);



protected:
	/*!
	 * Derive analytical solution with respect to i1 and i2
	 * @param i1 Workingset index
	 * @param i2 Workingset index
	 */
	void analyticSolve (int i1, int i2);

	inline Type l1Regularizer(Type tClass);
	inline Type l2Regularizer(Type tClass);

	/*!
	 * Update the gradient of the active items \f$ g_i \f$ using \f[ g(i) = g(i) + \Delta\alpha_a * K_{i,a} + \Delta\alpha_b * K_{i,b} \f]
	 * @param a Workingset index
	 * @param dDeltaA Alpha change of a
	 * @param b Workingset index
	 * @param dDeltaB Alpha change of b
	 **/
	void updateGradient (int& a, double& tDeltaA, int& b, double& tDeltaB);

	/*!
	 * Initialize algorithm with given dataset
	 * @param tDataset
	 **/
	void init();

	/*!
	 * Train the algorithm with the current trainset
	 **/
	void optimizeWorkingSet ();

	bool checkShrinking (int, int);

	/*!
	 * Recalculate the gradient
	 * \f[ g(i) = \sum_{j=1}^{N} y_i \alpha_i K_{i,j} - 1 \f]
	 * for all training examples.
	 **/
	void rebuildGradient();


	/*!
	 * Calculate bias for trained svm
	 * @return New bias value
	 **/
	Type calculateBias ();


	bool selectPair (int *piI1, int *piI2, Type & tError);




};

#endif
