/*
  COPYRIGHT (C) 2003 APPLIED NEUROINFORMATIC GROUP - UNIVERSITY OF BIELEFELD.

  ALL RIGHTS RESERVED.

  REDISTRIBUTION AND USE IN SOURCE AND BINARY FORM, WITH OR WITHOUT
  MODIFICATION, REQUIRE THE PERMISSION OF THE COPYRIGHT HOLDERS.

  COMMERCIAL USE WITHOUT THE EXPLICIT PERMISSION OF THE COPYRIGHT HOLDERS
  IS FORBIDDEN

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS IS''
  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE LIABLE FOR
  ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/



#ifndef CSPARSEVECTOR_H
#define CSPARSEVECTOR_H

#include <iostream>
#include <utility>

#include "CVector.h"
#include "keyed_iterator.hpp"



/*!
 * @brief Sparse, immutable vector representation.
 */

template <class Type>
class CSparseVector : public CVector<Type>
{
public:
	typedef pair<int, Type> element_type;
	typedef element_type* pointer_type;
	typedef keyed_iterator<Type> iterator;

private:
	pointer_type elements;
	int dim;
	int size;
	static Type nullElement;

	// forbidden
	CSparseVector& operator= (const CSparseVector &rhs);

public:
	CSparseVector(int dim, int size, Type elements[], int pos[]);
	CSparseVector(int dim, int size, vector<Type> elements, vector<int> pos);
	CSparseVector(int dim, int size);
	CSparseVector(const CSparseVector<Type> &o);
	CSparseVector(const CVector<Type>& o);
	CVector<Type>* clone() const;
	virtual ~CSparseVector();

	const Type& operator[] (int index) const;
	int dimension() const { return dim; }
	Type getElement(int iIndex) const;
	void setElement(int index, pair<int, Type> value);
	int getNumNonZero() const { return size; }
	bool isNaN() const;
	Type euclideanLength(bool bQuad=false) const;
	virtual string className() const { return string("CSparseVector");};
	CVector<double>* toDouble() const;
	CVector<float>* toFloat() const;
	CDenseVector<int> sort();

	CVector<Type>& operator*=(const CVector<Type>& v2) {
		CVector<Type>::operator*=(v2);
		return *this;
	}
	CVector<Type>& operator/=(const CVector<Type>& v2) {
		CVector<Type>::operator/=(v2);
		return *this;
	}
	/// multiply every element with scalar
	virtual CVector<Type>& operator*=(const Type&);
	/// divide every element by scalar
	virtual CVector<Type>& operator/=(const Type&);

	// fix to be <const Type>
	keyed_iterator<Type> begin() const {
		return keyed_iterator<Type>(elements, size, dim);
	}
	keyed_iterator<Type> begin() {
		return keyed_iterator<Type>(elements, size, dim);
	}
	keyed_iterator<Type> end() const {
		return keyed_iterator<Type>(elements + size, 0, dim);
	}
	keyed_iterator<Type> end() {
		return keyed_iterator<Type>(elements + size, 0, dim);
	}


};

// operator definitions

template<class Type>
std::ostream& operator << (std::ostream& stream, const CSparseVector<Type>& rhs)
{
	stream << "SparseVector[dim=" << rhs.dimension()
	       << ", size=" << rhs.getNumNonZero() << ", pairs={";

	for(typename CSparseVector<Type>::iterator i = rhs.begin();
	    i != rhs.end(); ++i) {
		stream << "(" << i.key() <<  ", " << i.val() << ") ";
	}
	stream << "}]";

	return stream;
}

template <class Type>
CSparseVector<Type> operator* (const CSparseVector<Type>& vec, const Type& s);
template <class Type>
CSparseVector<Type> operator* (const Type& s, const CSparseVector<Type> &vec);
template<class Type>
CSparseVector<Type> operator/ (const CSparseVector<Type>&, const Type&);
template<class Type>
CSparseVector<Type> elementMul(const CSparseVector<Type>&, const CVector<Type>&);

#endif // CSPARSEVECTOR_H
