/*
  COPYRIGHT (C) 2003 APPLIED NEUROINFORMATIC GROUP - UNIVERSITY OF BIELEFELD.

  ALL RIGHTS RESERVED.

  REDISTRIBUTION AND USE IN SOURCE AND BINARY FORM, WITH OR WITHOUT
  MODIFICATION, REQUIRE THE PERMISSION OF THE COPYRIGHT HOLDERS.

  COMMERCIAL USE WITHOUT THE EXPLICIT PERMISSION OF THE COPYRIGHT HOLDERS
  IS FORBIDDEN

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS IS''
  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE LIABLE FOR
  ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/


#ifndef CVECTOR_H
#define CVECTOR_H

#include <cassert>
#include <cstring>
#include <cstdio>
#include <cmath>
#include <iostream>
#include <vector>
#include "Macros.h"
#include "CException.h"
#include "CVectorOperations.h"
#include "CObject.h"
using std::vector;



template<class Type> struct keyed_iterator;
template<typename Type> class CDenseVector;
template<typename Type> class CSparseVector;

/**\class CVector
 *  @brief templatized vector for numerical applications
 *
 * base class for a generalized vector.  specifies all operations
 * that can be performed on a vector of any type.
 */
template <class Type>
class CVector : public CObject<Type>
{
protected:
	static const CVectorOperations<Type>& op_table;

public:
	virtual ~CVector() {};	///< destructor
	virtual CVector<Type>* clone() const = 0; ///< create identical copy

	/// create a copy with every element stored as 'double'
	virtual CVector<double>* toDouble() const = 0;
	/// create a copy with every element stored as 'float'
	virtual CVector<float>* toFloat() const = 0;

	virtual vector<Type> toSTL() const;
	virtual CDenseVector<int> sort() = 0;

	/// returns true if rhs has the same elements
	virtual bool operator==(const CVector<Type>& rhs) const;

	/// scalar product with vector
	Type operator*(const CVector<Type>&) const;
	/// multiply vectors elementwise
	CVector<Type>& operator*=(const CVector<Type>&);
	/// divide vectors elementwise
	CVector<Type>& operator/=(const CVector<Type>&);

	/// multiply every element with scalar
	virtual CVector<Type>& operator*=(const Type&) = 0;
	/// divide every element by scalar
	virtual CVector<Type>& operator/=(const Type&) = 0;

	/// return vector dimension
	virtual int dimension(void) const = 0;
	/// returns the number of non-zero entries
	virtual int getNumNonZero() const = 0;

	/// returns true if vector contains nan elements
	virtual bool isNaN() const = 0;

	/// returns euclidean length of vector
	virtual Type euclideanLength(bool bQuad=false) const = 0;

	/// return constant reference to n'th vector element
	virtual const Type& operator[](int n) const = 0;
	/// return copy of n'th vector element
	virtual Type getElement(int n) const = 0;
	/*!
	**/
	virtual string className() const { return string("CVector"); };

	/** euclidean distance of the two vectors
	 * @param bQuad true - return sum of squares; false - return sqrt of that
	 */
	Type euclideanDistance(const CVector<Type>& , bool b=false)  const;

	/// returns common cosine distance of the two vectors
	Type coCosineDistance(const CVector<Type>&) const;



	/** adapt vector according to
	 *  v1 = v1 + eps * (v2-v1)
	 *
	 *  since v2 might be sparse, we achieve a major speedup, if we
	 *  formulate the adaptation rule slightly different:
	 *  v1 = eps2 * v1 + eps1 * v2,  where eps2 = 1-eps1
	 *
	 * @param v2 - vector to adapt to
	 * @param fEps1 - eps1
	 * @param fEps2 - eps2
	 */
	void adaptToVec(const CVector<Type>&, float fEps1, float fEps2);

	virtual  bool serialize(fstream& stream, IO_MODE mode);

}; // end of class


/*
 * Local variables:
 * mode: c++
 * End:
 */

#endif
