/*
  COPYRIGHT (C) 2003 APPLIED NEUROINFORMATIC GROUP - UNIVERSITY OF BIELEFELD.

  ALL RIGHTS RESERVED.

  REDISTRIBUTION AND USE IN SOURCE AND BINARY FORM, WITH OR WITHOUT
  MODIFICATION, REQUIRE THE PERMISSION OF THE COPYRIGHT HOLDERS.

  COMMERCIAL USE WITHOUT THE EXPLICIT PERMISSION OF THE COPYRIGHT HOLDERS
  IS FORBIDDEN

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS IS''
  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE LIABLE FOR
  ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/



#ifndef __SPARSE_ALGORITHM_H
#define __SPARSE_ALGORITHM_H

#include <numeric>

// specializations of the STL numerical algorithms for
// work with sparse iterators

// BE CAREFUL: 'inner_product' expects its BinaryFunctions to have the
// properties of plus and multiplies with regard to an argument or an
// intermediate result being 0

namespace std {
template<typename Iterator, typename Type, typename BinaryFunction1, typename BinaryFunction2>
Type inner_product(Iterator lhs, Iterator lhs_end,
                   keyed_iterator<Type> rhs, Type init,
                   BinaryFunction1 f1, BinaryFunction2 f2) {
	Type result = init;
	keyed_iterator<Type> rhs_end = rhs + rhs.size();

	while(rhs != rhs_end) {
		result = f1(result, f2(lhs[rhs.key()], rhs.val()));
		++rhs;
	}
	return result;
}

template<typename Iterator, typename Type, typename BinaryFunction1, typename BinaryFunction2>
Type inner_product(keyed_iterator<Type> lhs, keyed_iterator<Type> lhs_end,
                   Iterator rhs, Type init,
                   BinaryFunction1 f1, BinaryFunction2 f2) {
	Type result = init;

	while (lhs != lhs_end) {
		result = f1(result, f2(lhs.val(), rhs[lhs.key()]));
		++lhs;
	}
	return result;
}

template<typename Type, typename BinaryFunction1, typename BinaryFunction2>
Type inner_product(keyed_iterator<Type> lhs, keyed_iterator<Type> lhs_end,
                   keyed_iterator<Type> rhs, Type init,
                   BinaryFunction1 f1, BinaryFunction2 f2)
{
	Type result = init;
	keyed_iterator<Type> rhs_end = rhs + rhs.size();

	while (lhs != lhs_end && rhs != rhs_end) {

		if (lhs.key() < rhs.key() && lhs != lhs_end) {
			lhs++;
		} else if (lhs.key() == rhs.key()) {
			result = f1(result, f2(lhs.val(), rhs.val()));
			lhs++;
			rhs++;
		} else if(lhs.key() > rhs.key()) {
			rhs++;
		}
	}

	return result;
}

} // namespace std



template<typename Iterator, typename Type,
         typename BinaryFunction1, typename BinaryFunction2>
Type inner_distance(Iterator lhs, Iterator lhs_end,
                    keyed_iterator<Type> rhs, Type init,
                    BinaryFunction1 f1, BinaryFunction2 f2) {
	Type result = init;
	keyed_iterator<Type> rhs_end = rhs + rhs.size();
	int pos = 0;

	while(rhs != rhs_end) {
		int rhs_pos = rhs.key();

		for(; pos < rhs_pos; ++pos, ++lhs) {
			result = f1(result, f2(*lhs, static_cast<Type>(0)));
		}
		result = f1(result, f2(*lhs, rhs.val()));

		++lhs;
		++rhs;
		++pos;
	}

	// iterate over the rest
	while(lhs != lhs_end) {
		result = f1(result, f2(*lhs, static_cast<Type>(0)));
		++lhs;
	}

	return result;
}

template<typename Iterator, typename Type,
         typename BinaryFunction1, typename BinaryFunction2>
Type inner_distance(keyed_iterator<Type> lhs, keyed_iterator<Type> lhs_end,
                    Iterator rhs, Type init,
                    BinaryFunction1 f1, BinaryFunction2 f2)
{
	Type result = init;
	Iterator rhs_end = rhs + lhs.max();
	int pos = 0;

	while(lhs != lhs_end) {
		int lhs_pos = lhs.key();

		for(; pos < lhs_pos; ++pos, ++rhs) {
			result = f1(result, f2(static_cast<Type>(0), *rhs));
		}
		result = f1(result, f2(lhs.val(), *rhs));

		++lhs;
		++rhs;
		++pos;
	}

	// when dense iterator has entries beyond last entry in
	// sparse iterator
	while(rhs != rhs_end) {
		result = f1(result, f2(static_cast<Type>(0), *rhs));
		++rhs;
	}

	return result;
}

template<typename Type,
         typename BinaryFunction1, typename BinaryFunction2>
Type inner_distance(keyed_iterator<Type> lhs, keyed_iterator<Type> lhs_end,
                    keyed_iterator<Type> rhs, Type init,
                    BinaryFunction1 f1, BinaryFunction2 f2) {
	Type result = init;
	keyed_iterator<Type> rhs_end = rhs + rhs.size();

	while(lhs != lhs_end && rhs != rhs_end) {
		if(lhs.key() < rhs.key()) {
			result = f1(result, f2(lhs.val(), static_cast<Type>(0)));
			++lhs;
		} else if(lhs.key() == rhs.key()) {
			result = f1(result, f2(lhs.val(), rhs.val()));
			++lhs;
			++rhs;
		} else if(lhs.key() > rhs.key()) {
			result = f1(result, f2(static_cast<Type>(0), rhs.val()));
			++rhs;
		}
	}
	// consume leftovers
	while(lhs != lhs_end) {
		result = f1(result, f2(lhs.val(), static_cast<Type>(0)));
		++lhs;
	}
	while(rhs != rhs_end) {
		result = f1(result, f2(static_cast<Type>(0), rhs.val()));
		++rhs;
	}

	return result;
}


// fall back for standard iterators
template<typename Iterator1, typename Iterator2, typename Type,
         typename BinaryFunction1, typename BinaryFunction2>
Type inner_distance(Iterator1 lhs, Iterator1 lhs_end,
                    Iterator2 rhs, Type init,
                    BinaryFunction1 f1, BinaryFunction2 f2) {
	return inner_product(lhs, lhs_end, rhs, init, f1, f2);
}


// CO COSINE DISTANCE

template<typename Iterator, typename Type, typename BinaryFunction1, typename BinaryFunction2>
Type coCosine_distance(Iterator lhs, Iterator lhs_end,
                       keyed_iterator<Type> rhs, Type init,
                       BinaryFunction1 f1, BinaryFunction2 f2) {

	Type result = init;
	Type lLen = init;
	Type rLen = init;

	keyed_iterator<Type> rhs_end = rhs + rhs.size();

	while (rhs != rhs_end) {
		result = f1(result, f2(lhs[rhs.key()], rhs.val()));
		lLen = f1(lLen, f2(lhs[rhs.key()], rhs.val()));
		rLen += rhs.val() * rhs.val();
		rhs++;
	}

	if (lLen * rLen > 0) result /= sqrt(lLen * rLen);
	return result;
}

template<typename Iterator, typename Type, typename BinaryFunction1, typename BinaryFunction2>
Type coCosine_distance(keyed_iterator<Type> lhs, keyed_iterator<Type> lhs_end,
                       Iterator rhs, Type init,
                       BinaryFunction1 f1, BinaryFunction2 f2) {

	Type result = init;
	Type lLen = 0;
	Type rLen = 0;

	while (lhs != lhs_end) {
		result = f1(result, f2(lhs.val(), rhs[lhs.key()]));
		lLen += lhs.val() * lhs.val();
		rLen += rhs[lhs.key()] * rhs[lhs.key()];
		lhs++;
	}

	if (lLen * rLen > 0) result /= sqrt(lLen * rLen);
	return result;
}


template<typename Type, typename BinaryFunction1, typename BinaryFunction2>
Type coCosine_distance(keyed_iterator<Type> lhs, keyed_iterator<Type> lhs_end,
                       keyed_iterator<Type> rhs, Type init,
                       BinaryFunction1 f1, BinaryFunction2 f2) {

	Type result = init;
	Type lLen = 0;
	Type rLen = 0;

	keyed_iterator<Type> rhs_end = rhs + rhs.size();

	while (lhs != lhs_end && rhs != rhs_end) {

		if (lhs.key() < rhs.key() && lhs != lhs_end) {
			lhs++;
		} else if (lhs.key() == rhs.key()) {
			result = f1(result, f2(lhs.val(), rhs.val()));
			lLen += lhs.val() * lhs.val();
			rLen += rhs.val() * rhs.val();

			lhs++;
			rhs++;
		} else if(lhs.key() > rhs.key()) {
			rhs++;
		}
	}

	if (lLen * rLen > 0) result /= sqrt(lLen * rLen);
	return result;
}


// fall back for standard iterators
template<typename Iterator1, typename Iterator2, typename Type, typename BinaryFunction1, typename BinaryFunction2>
Type coCosine_distance(Iterator1 lhs, Iterator1 lhs_end,
                       Iterator2 rhs, Type init,
                       BinaryFunction1 f1, BinaryFunction2 f2) {

	Type result = init;
	Type lLen = 0;
	Type rLen = 0;

	while (lhs != lhs_end) {
		result = f1(result, f2(*lhs, *rhs));
		lLen += *lhs * *lhs;
		rLen += *rhs * *rhs;
		lhs++;
		rhs++;
	}

	if (lLen * rLen > 0) result /= sqrt(lLen * rLen);
	return result;
}

// ELEMENT_PRODUCT

template<typename T1, typename BinaryFunction>
void element_product(keyed_iterator<T1> l, keyed_iterator<T1> l_end,
                     keyed_iterator<T1> r, keyed_iterator<T1> r_end,
                     keyed_iterator<T1> out,
                     BinaryFunction f)
{
	for(; l != l_end; ++l, ++out) {
		while(r != r_end && r.key() < l.key()) {
			++r;
		}
		if(r != r_end && r.key() == l.key()) {
			out.val() = f(l.val(), r.val());
		} else {
			out.val() = 0;
		}
	}
}

template<typename Type, typename BinaryFunction, typename Iterator>
void element_product(keyed_iterator<Type> l, keyed_iterator<Type> l_end,
                     keyed_iterator<Type> r, keyed_iterator<Type> r_end,
                     Iterator out, BinaryFunction f)
{
	// because we might not hit every index, delete it first
	memset(out, 0,  l.max() * sizeof(Type));

	for(; l != l_end; ++l, ++r) {
		while(r != r_end && r.key() < l.key()) {
			++r;
		}
		if(r != r_end && r.key() == l.key()) {
			out[r.key()] = f(l.val(), r.val());
		}
	}
}


template<typename Iterator, typename Ti,
         typename To, typename BinaryFunction>
void element_product(Iterator l, Iterator l_end,
                     keyed_iterator<Ti> r, keyed_iterator<Ti> r_end,
                     keyed_iterator<To> out,
                     BinaryFunction f)
{
	for(int i = 0; l != l_end; ++i, ++l) {
		if(i == r.key()) {
			out.val() = f(*l, r.val());
			++r;
			++out;
		}
	}
}

template<typename Iterator, typename Type, typename BinaryFunction,
         typename OutputIterator>
void element_product(Iterator l, Iterator l_end,
                     keyed_iterator<Type> r, keyed_iterator<Type> r_end,
                     OutputIterator out, BinaryFunction f)
{
	for(int i = 0; l != l_end; ++i, ++l, ++out) {
		if(i == r.key()) {
			*out = f(*l, r.val());
			++r;
		} else {
			*out = f(*l, 0);
		}
	}

}

template<typename Type, typename Iterator, typename BinaryFunction>
void element_product(keyed_iterator<Type> l, keyed_iterator<Type> l_end,
                     Iterator r, Iterator r_end,
                     keyed_iterator<Type> out, BinaryFunction f)
{
	while(l != l_end) {
		*out = make_pair(l.key(), f(l.val(), r[l.key()]));
		++l; ++out;
	}
}

template<typename Type, typename Iterator, typename BinaryFunction, typename OutputIterator>
void element_product(keyed_iterator<Type> l, keyed_iterator<Type> l_end,
                     Iterator r, Iterator r_end,
                     OutputIterator out, BinaryFunction f)
{
	int index = 0;
	while(r != r_end) {
		if(index == l.key()) {
			*out = f(l.val(), *r);
			++l;
		}
		++out;
		++r;
		++index;
	}
}



template<typename Iterator1, typename Iterator2,
         typename OutputIterator, typename BinaryFunction>
void element_product(Iterator1 l, Iterator1 l_end,
                     Iterator2 r, Iterator2 r_end,
                     OutputIterator out, BinaryFunction f)
{
	while(l != l_end) {
		*out = f(*l, *r);
		++l; ++r; ++out;
	}
}

/*template<typename Iterator, typename Type, typename BinaryFunction>
  void element_product(Iterator l, Iterator l_end, Type val,
  keyed_iterator<Type> out, BinaryFunction f)
  {
  for(; l != l_end; ++l, ++out) {
  out.val() = f(*l, val);
  }
  }*/

template<typename T1, typename BinaryFunction>
void element_product(keyed_iterator<T1> l, keyed_iterator<T1> l_end,
                     T1 val, keyed_iterator<T1> out, BinaryFunction f)
{
	for(; l != l_end; ++l, ++out) {
		out.val() = f(l.val(), val);
	}
}

template<typename Iterator, typename Type, typename BinaryFunction>
void element_product(Iterator l, Iterator l_end, Type val, Iterator out,
                     BinaryFunction f)
{
	while(l != l_end) {
		*out = f(*l, val);
		++l; ++ out;
	}
}

// ELEMENT_ADD

template<typename Iterator, typename OutputIterator, typename BinaryFunction>
void element_add(Iterator l, Iterator l_end,
                 Iterator r, Iterator r_end,
                 OutputIterator out, BinaryFunction f)
{
	while(l != l_end) {
		*out = f(*l, *r);
		++l; ++r; ++out;
	}
}
template<typename Type, typename OutputIterator, typename BinaryFunction>
void element_add(keyed_iterator<Type> l, keyed_iterator<Type> l_end,
                 keyed_iterator<Type> r, keyed_iterator<Type> r_end,
                 OutputIterator out, BinaryFunction f)
{
	memset(out, 0, l.max() * sizeof(Type));
	while(l != l_end) {
		out[l.key()] = l.val();
		++l;
	}
	while(r != r_end) {
		out[r.key()] = f(out[r.key()], r.val());
		++r;
	}
}
template<typename Iterator, typename Type,
         typename OutputIterator, typename BinaryFunction>
void element_add(Iterator l, Iterator l_end,
                 keyed_iterator<Type> r, keyed_iterator<Type> r_end,
                 OutputIterator out, BinaryFunction f)
{
	OutputIterator start = out;

	while(l != l_end) {
		*out = *l;
		++l; ++out;
	}
	while(r != r_end) {
		start[r.key()] = f(start[r.key()], r.val());
		++r;
	}
}
template<typename Iterator, typename Type,
         typename OutputIterator, typename BinaryFunction>
void element_add(keyed_iterator<Type> l, keyed_iterator<Type> l_end,
                 Iterator r, Iterator r_end,
                 OutputIterator out, BinaryFunction f)
{
	int pos = 0;

	while(r != r_end) {
		if(pos == l.key()) {
			*out = f(l.val(), *r);
			++l;
		} else {
			*out = f(static_cast<Type>(0), *r);
		}
		++r; ++out; ++pos;
	}
}

// used for vector adaption ----------------------------------------
template<typename Type, typename BinaryFunction>
void element_adapt(keyed_iterator<Type> l, keyed_iterator<Type> l_end,
                   keyed_iterator<Type> r, keyed_iterator<Type> r_end,
                   keyed_iterator<Type> out,
                   BinaryFunction f, float fEps1, float fEps2)
{
	throw CException("Adaptation of sparse vectors is not supported.");
}

template<typename Type, typename Iterator, typename BinaryFunction, typename OutputIterator>
void element_adapt(keyed_iterator<Type> l, keyed_iterator<Type> l_end,
                   Iterator r, Iterator r_end,
                   OutputIterator out, BinaryFunction f, float fEps1, float fEps2)
{
	throw CException("Adaptation of sparse vectors is not supported.");
}

template<typename Iterator, typename Type, typename BinaryFunction, typename OutputIterator>
void element_adapt(Iterator l, Iterator l_end,
                   keyed_iterator<Type> r, keyed_iterator<Type> r_end,
                   OutputIterator out, BinaryFunction f, float fEps1, float fEps2)
{
	OutputIterator start = out;

	if (fEps2 != 1) {
		while (l != l_end) {
			*out = fEps2 * *l;
			++l; ++out;
		}
	}

	while (r != r_end) {
		start[r.key()] = f(start[r.key()], fEps1 * r.val());
		++r;
	}
}

template<typename Iterator1, typename Iterator2, typename OutputIterator, typename BinaryFunction>
void element_adapt(Iterator1 l, Iterator1 l_end,
                   Iterator2 r, Iterator2 r_end,
                   OutputIterator out, BinaryFunction f, float fEps1, float fEps2)
{
	if (fEps2 == 1) {
		while (l != l_end) {
			*out = f(*l, fEps1 * *r);
			++l; ++out; ++r;
		}
	} else {
		while (l != l_end) {
			*out = f(fEps2 * *l, fEps1 * *r);
			++l; ++out; ++r;
		}
	}
}


template<typename Type>
struct euclidean_norm {
	Type operator()(const Type v1, const Type v2) const
	{
		return (v2 - v1)*(v2 - v1);
	}
};

/*
 * Local variables:
 * mode: c++
 * End:
 */

#endif // #ifndef __SPARSE_ALGORITHM_H
